
##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
# See CREDITS file for contributors and development history
#
##############################################################################


__copyright__="""Copyright (c) 2003-2026 by the esys.escript Group
https://github.com/LutzGross/esys-escript.github.io
Primary Business: Queensland, Australia"""
__license__="""Licensed under the Apache License, version 2.0
http://www.apache.org/licenses/LICENSE-2.0"""
__url__="https://github.com/LutzGross/esys-escript.github.io"


from .finleycpp import __Brick_driver, __Rectangle_driver, __ReadMesh_driver, __ReadGmsh_driver


def ReadMesh(filename, integrationOrder=-1, reducedIntegrationOrder=-1, optimize=True,
             diracPoints=[], diracTags=[], comm=None, framework=None):
    """
    Read a mesh from a file. For MPI parallel runs fan out the mesh to multiple processes.

    :param filename: Path to mesh file
    :param integrationOrder: Order of quadrature scheme (-1 for automatic)
    :param reducedIntegrationOrder: Order of reduced quadrature scheme (-1 for automatic)
    :param optimize: Enable optimization of node labels
    :param diracPoints: Dirac point coordinates
    :param diracTags: Dirac point tags
    :param comm: MPI communicator (optional, defaults to MPI_COMM_WORLD)
    :param framework: solver framework to use (optional SolverFramework instance)
    :return: Domain object
    """
    dom = __ReadMesh_driver(filename, integrationOrder, reducedIntegrationOrder,
                            optimize, diracPoints, diracTags, comm)
    if framework is not None:
        dom.setFramework(framework)
    return dom

def ReadGmsh(fileName, numDim, integrationOrder=-1, reducedIntegrationOrder=-1, optimize=True,
             useMacroElements=False, diracPoints=[], diracTags=[], comm=None, framework=None):
    """
    Read a gmsh mesh file.

    :param fileName: Path to gmsh file
    :param numDim: Number of spatial dimensions
    :param integrationOrder: Order of quadrature scheme (-1 for automatic)
    :param reducedIntegrationOrder: Order of reduced quadrature scheme (-1 for automatic)
    :param optimize: Enable optimization of node labels
    :param useMacroElements: Enable usage of macro elements instead of second order elements
    :param diracPoints: Dirac point coordinates
    :param diracTags: Dirac point tags
    :param comm: MPI communicator (optional, defaults to MPI_COMM_WORLD)
    :param framework: solver framework to use (optional SolverFramework instance)
    :return: Domain object
    """
    dom = __ReadGmsh_driver(fileName, numDim, integrationOrder, reducedIntegrationOrder,
                            optimize, useMacroElements, diracPoints, diracTags, comm)
    if framework is not None:
        dom.setFramework(framework)
    return dom


def Rectangle(n0=1, n1=1, order=1, l0=1.0, l1=1.0, periodic0=False, periodic1=False, integrationOrder=-1,
               reducedIntegrationOrder=-1, useElementsOnFace=None, useFullElementOrder=False, optimize=False,
               diracPoints=[], diracTags=[], comm=None, framework=None):
    """
    Creates a rectangular mesh with n0 x n1 elements over the rectangle [0,l0] x [0,l1].

    :param n0: number of elements in direction 0
    :param n1: number of elements in direction 1
    :param order: order of shape functions (1, 2, or -1 for macro elements)
    :param l0: length of side 0
    :param l1: length of side 1
    :param periodic0: periodic boundary conditions in direction 0
    :param periodic1: periodic boundary conditions in direction 1
    :param integrationOrder: order of quadrature scheme (-1 for automatic)
    :param reducedIntegrationOrder: order of reduced quadrature scheme (-1 for automatic)
    :param useElementsOnFace: whether to use elements on face
    :param useFullElementOrder: whether to use Rec9 elements
    :param optimize: enable optimization of node labels
    :param diracPoints: Dirac point coordinates
    :param diracTags: Dirac point tags
    :param comm: MPI communicator (optional, defaults to MPI_COMM_WORLD)
    :param framework: solver framework to use (optional SolverFramework instance)
    :return: Domain object
    """
    faceon = useElementsOnFace
    if useElementsOnFace is None:  # We want to use True as the default, but only where it makes sense
        if useFullElementOrder or order == -1:
            faceon = False  # Don't use it
        else:
            faceon = True

    dom = __Rectangle_driver((n0, n1), order, (l0, l1), (periodic0, periodic1),
                             integrationOrder, reducedIntegrationOrder,
                             faceon, useFullElementOrder, optimize,
                             diracPoints, diracTags, comm)
    if framework is not None:
        dom.setFramework(framework)
    return dom

def Brick(n0=1, n1=1, n2=1, order=1, l0=1.0, l1=1.0, l2=1.0, periodic0=False, periodic1=False, periodic2=False,
          integrationOrder=-1, reducedIntegrationOrder=-1, useElementsOnFace=None, useFullElementOrder=False,
          optimize=False, diracPoints=[], diracTags=[], comm=None, framework=None):
    """
    Creates a rectangular mesh with n0 x n1 x n2 elements over the brick [0,l0] x [0,l1] x [0,l2].

    :param n0: number of elements in direction 0
    :param n1: number of elements in direction 1
    :param n2: number of elements in direction 2
    :param order: order of shape functions (1, 2, or -1 for macro elements)
    :param l0: length of side 0
    :param l1: length of side 1
    :param l2: length of side 2
    :param periodic0: periodic boundary conditions in direction 0
    :param periodic1: periodic boundary conditions in direction 1
    :param periodic2: periodic boundary conditions in direction 2
    :param integrationOrder: order of quadrature scheme (-1 for automatic)
    :param reducedIntegrationOrder: order of reduced quadrature scheme (-1 for automatic)
    :param useElementsOnFace: whether to use elements on face
    :param useFullElementOrder: whether to use Hex27 elements
    :param optimize: enable optimization of node labels
    :param diracPoints: Dirac point coordinates
    :param diracTags: Dirac point tags
    :param comm: MPI communicator (optional, defaults to MPI_COMM_WORLD)
    :param framework: solver framework to use (optional SolverFramework instance)
    :return: Domain object
    """
    faceon = useElementsOnFace
    if useElementsOnFace is None:  # We want to use True as the default, but only where it makes sense
        if useFullElementOrder or order == -1:
            faceon = False  # Don't use it
        else:
            faceon = True

    dom = __Brick_driver((n0, n1, n2), order, (l0, l1, l2), (periodic0, periodic1, periodic2),
                         integrationOrder, reducedIntegrationOrder,
                         faceon, useFullElementOrder, optimize,
                         diracPoints, diracTags, comm)
    if framework is not None:
        dom.setFramework(framework)
    return dom




##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
# See CREDITS file for contributors and development history
#
##############################################################################

"""
Utility functions for symbolic computations in escript.

This module provides helper functions for working with symbolic expressions,
including creating Symbol objects and computing total differentials.

:var __author__: name of author
:var __copyright__: copyrights
:var __license__: licence agreement
:var __url__: url entry point on documentation
:var __version__: version
:var __date__: date of the version
"""


__copyright__="""Copyright (c) 2003-2026 by the esys.escript Group
https://github.com/LutzGross/esys-escript.github.io
Primary Business: Queensland, Australia"""
__license__="""Licensed under the Apache License, version 2.0
http://www.apache.org/licenses/LICENSE-2.0"""
__url__="https://github.com/LutzGross/esys-escript.github.io"
__author__="Cihan Altinay"

import numpy
import sympy
from .symbol import Symbol

def symbols(*names, **kwargs):
    """
    Creates one or more Symbol objects, emulating sympy.symbols behavior.

    :param names: symbol name(s), space or comma separated
    :type names: ``str``
    :keyword shape: shape of the symbol (default: scalar)
    :type shape: ``tuple`` of ``int``
    :keyword dim: spatial dimension
    :type dim: ``int``
    :return: single Symbol or tuple of Symbols
    :rtype: `Symbol` or ``tuple`` of `Symbol`
    """

    shape=kwargs.pop('shape', ())

    s = names[0]
    if not isinstance(s, list):
        import re
        s = re.split(r'\s|,', s)
    res = []
    for t in s:
        # skip empty strings
        if not t:
            continue
        sym = Symbol(t, shape, **kwargs)
        res.append(sym)
    res = tuple(res)
    if len(res) == 0:   # var('')
        res = None
    elif len(res) == 1: # var('x')
        res = res[0]
                        # otherwise var('a b ...')
    return res


def isSymbol(arg):
    """
    Tests if the argument is a symbolic object.

    :param arg: object to test
    :type arg: any
    :return: ``True`` if ``arg`` is an escript `Symbol` or ``sympy.Basic``
             object, ``False`` otherwise
    :rtype: ``bool``
    """
    return isinstance(arg, Symbol) or isinstance(arg, sympy.Basic)

def removeFsFromGrad(sym):
    """
    Removes function space parameters from gradient expressions.

    Returns ``sym`` with all occurrences of grad_n(a,b,c) replaced by
    grad_n(a,b). This strips the function space parameter from gradient calls.

    :param sym: symbolic expression to process
    :type sym: `Symbol`
    :return: expression with function space parameters removed from gradients
    :rtype: `Symbol`
    """
    from esys.escript import symfn
    gg=sym.atoms(symfn.grad_n)
    for g in gg:
        if len(g.args)==3:
            r=symfn.grad_n(*g.args[:2])
            sym=sym.subs(g, r)
    return sym

def getTotalDifferential(f, x, order=0):
    """
    Computes the total differential of f with respect to x.

    This function computes::

        | Df/Dx = del_f/del_x + del_f/del_grad(x)*del_grad(x)/del_x + ...
        |            \\   /         \\   /
        |              a             b

    :param f: the function to differentiate
    :type f: `Symbol` or ``numpy.ndarray``
    :param x: the variable to differentiate with respect to
    :type x: `Symbol`
    :param order: order of differentiation (0 or 1)
    :type order: ``int``
    :return: tuple of derivative terms (a,) for order=0, (a, b) for order=1
    :rtype: `Symbol` or ``tuple`` of `Symbol`
    """

    from .. import util
    res=()
    shape=util.getShape(f)
    if not isSymbol(f):
        res+=(numpy.zeros(shape+x.getShape()),)
        for i in range(order):
            x=x.grad()
            res+=numpy.zeros(shape+x.getShape())

    elif x.getRank()==0:
        f=removeFsFromGrad(f)
        dfdx=f.diff(x)
        dgdx=x.grad().diff(x)
        a=numpy.empty(shape, dtype=object)
        if order>0:
            b=numpy.empty(shape+dgdx.getShape(), dtype=object)

        if len(shape)==0:
            for j in numpy.ndindex(dgdx.getShape()):
                y=dfdx
                z=dgdx[j]
                # expand() and coeff() are very expensive so
                # we set the unwanted factors to zero to extract
                # the one we need
                for jj in numpy.ndindex(dgdx.getShape()):
                    if j==jj: continue
                    y=y.subs(dgdx[jj], 0)
                a=y.subs(z,0) # terms in x and constants
                if order>0:
                    b[j]=y.subs(z,1)-a
        else:
            for i in numpy.ndindex(shape):
                for j in numpy.ndindex(dgdx.getShape()):
                    y=dfdx[i]
                    z=dgdx[j]
                    for jj in numpy.ndindex(dgdx.getShape()):
                        if j==jj: continue
                        y=y.subs(dgdx[jj], 0)
                    a[i]=y.subs(z,0) # terms in x and constants
                    if order>0:
                        b[i+j]=y.subs(z,1)-a[i]
        res+=(Symbol(a, dim=f.getDim(), subs=f.getDataSubstitutions()),)
        if order>0:
            res+=(Symbol(b, dim=f.getDim(), subs=f.getDataSubstitutions()),)

    elif x.getRank()==1:
        f=removeFsFromGrad(f)
        dfdx=f.diff(x)
        dgdx=x.grad().diff(x).transpose(2)
        a=numpy.empty(shape+x.getShape(), dtype=object)
        if order>0:
            b=numpy.empty(shape+x.grad().getShape(), dtype=object)

        if len(shape)==0:
            raise NotImplementedError('f scalar, x vector')
        else:
            for i in numpy.ndindex(shape):
                for k,l in numpy.ndindex(x.grad().getShape()):
                    if dgdx[k,k,l]==0:
                        a[i+(k,)]=0
                        if order>0:
                            b[i+(k,l)]=0
                    else:
                        y=dfdx[i+(k,)]
                        z=dgdx[k,k,l]
                        for kk,ll in numpy.ndindex(x.grad().getShape()):
                            if k==kk and l==ll: continue
                            y=y.subs(dgdx[kk,kk,ll], 0)
                        a[i+(k,)]=y.subs(z,0) # terms in x and constants
                        if order>0:
                            b[i+(k,l)]=y.subs(z,1)-a[i+(k,)]

        res+=(Symbol(a, dim=f.getDim(), subs=f.getDataSubstitutions()),)
        if order>0:
            res+=(Symbol(b, dim=f.getDim(), subs=f.getDataSubstitutions()),)

    if len(res)==1:
        return res[0]
    else:
        return res




##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
# See CREDITS file for contributors and development history
#
##############################################################################

# No-MPI configuration for Fedora Linux.
# Imports the full fedora_options.py and disables MPI.
#
# Prerequisites (MPI packages not required):
#   sudo dnf install python3-devel python3-numpy python3-scipy python3-sympy python3-matplotlib
#   sudo dnf install gcc-c++ gcc-gfortran scons cmake
#   sudo dnf install boost-devel boost-python3-devel boost-python3 boost-numpy3
#   sudo dnf install hdf5-devel netcdf-devel suitesparse-devel lapack-devel zlib-devel metis-devel

from scons.templates.fedora_options import *

# Disable MPI
mpi = 'none'
mpi4py = False




##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
# See CREDITS file for contributors and development history
#
##############################################################################

"""A domain meshed with uniform rectangles or quadrilaterals
"""


__copyright__="""Copyright (c) 2003-2026 by the esys.escript Group
https://github.com/LutzGross/esys-escript.github.io
Primary Business: Queensland, Australia"""
__license__="""Licensed under the Apache License, version 2.0
http://www.apache.org/licenses/LICENSE-2.0"""
__url__="https://github.com/LutzGross/esys-escript.github.io"

import esys.escript       # This is just to ensure required libraries are loaded
from .ripleycpp import *
from .ripleycpp import Rectangle as _Rectangle, Brick as _Brick
from .MultiResolutionDomain import *


def Rectangle(n0=10, n1=10, l0=1.0, l1=1.0, d0=-1, d1=-1,
              diracPoints=[], diracTags=[], comm=None, framework=None):
    """
    Creates a rectangular mesh with n0 x n1 elements over the rectangle [0,l0] x [0,l1].

    :param n0: number of elements in direction 0
    :param n1: number of elements in direction 1
    :param l0: length of side 0 or coordinate range of side 0
    :param l1: length of side 1 or coordinate range of side 1
    :param d0: number of subdivisions in direction 0
    :param d1: number of subdivisions in direction 1
    :param diracPoints: Dirac point coordinates
    :param diracTags: Dirac point tags
    :param comm: MPI communicator (optional, from mpi4py)
    :param framework: solver framework to use (optional SolverFramework instance)
    :return: Domain object
    """
    dom = _Rectangle(n0=n0, n1=n1, l0=l0, l1=l1, d0=d0, d1=d1,
                     diracPoints=diracPoints, diracTags=diracTags, comm=comm)
    if framework is not None:
        dom.setFramework(framework)
    return dom


def Brick(n0=10, n1=10, n2=10, l0=1.0, l1=1.0, l2=1.0, d0=-1, d1=-1, d2=-1,
          diracPoints=[], diracTags=[], comm=None, framework=None):
    """
    Creates a hexagonal mesh with n0 x n1 x n2 elements over the brick [0,l0] x [0,l1] x [0,l2].

    :param n0: number of elements in direction 0
    :param n1: number of elements in direction 1
    :param n2: number of elements in direction 2
    :param l0: length of side 0 or coordinate range of side 0
    :param l1: length of side 1 or coordinate range of side 1
    :param l2: length of side 2 or coordinate range of side 2
    :param d0: number of subdivisions in direction 0
    :param d1: number of subdivisions in direction 1
    :param d2: number of subdivisions in direction 2
    :param diracPoints: Dirac point coordinates
    :param diracTags: Dirac point tags
    :param comm: MPI communicator (optional, from mpi4py)
    :param framework: solver framework to use (optional SolverFramework instance)
    :return: Domain object
    """
    dom = _Brick(n0=n0, n1=n1, n2=n2, l0=l0, l1=l1, l2=l2, d0=d0, d1=d1, d2=d2,
                 diracPoints=diracPoints, diracTags=diracTags, comm=comm)
    if framework is not None:
        dom.setFramework(framework)
    return dom


__nodocorecursion=['ripleycpp']



# -*- coding: utf-8 -*-

##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
# See CREDITS file for contributors and development history
#
##############################################################################


__copyright__="""Copyright (c) 2003-2026 by the esys.escript Group
https://github.com/LutzGross/esys-escript.github.io
Primary Business: Queensland, Australia"""
__license__="""Licensed under the Apache License, version 2.0
http://www.apache.org/licenses/LICENSE-2.0"""
__url__="https://github.com/LutzGross/esys-escript.github.io"

"""
some mesh handling

:var __author__: name of author
:var __licence__: licence agreement
:var __url__: url entry point on documentation
:var __version__: version
:var __date__: date of the version
"""

__author__="Lutz Gross, l.gross@uq.edu.au"

from .factorywrappers import ReadGmsh, ReadMesh
from .finleycpp import LoadMesh


def GetMeshFromFile(filename, **kwargs):
    """
    Reads a mesh from a file, determines the reader to use based on the file
    extension. All cases require a filename and gmsh files require a number
    of dimensions (it doesn't hurt to pass this in all the time). Other
    keyword args come from the underlying reader functions.
    """
    spl=filename.split('.')
    ext=spl[len(spl)-1]
    # extract possible params
    integrationOrder=-1
    if "integrationOrder" in kwargs:
        integrationOrder=kwargs["integrationOrder"]
    reducedIntegrationOrder=-1
    if "reducedIntegrationOrder" in kwargs:
        integrationOrder=kwargs["reducedIntegrationOrder"]
    optimize=True  
    if "optimize" in kwargs:
        integrationOrder=kwargs["optimize"]
    useMacroElements=False
    if "useMacroElements" in kwargs:
        integrationOrder=kwargs["useMacroElements"]
    points = []
    if 'diracPoints' in kwargs:
        points=kwargs['diracPoints']
    tags = [ ]
    if 'diracTags' in kwargs:
        tags=kwargs['diracTags']

    if ext=="fly":
        return ReadMesh(filename, integrationOrder, reducedIntegrationOrder, optimize,  diracPoints=points, diracTags=tags)
    elif ext=="msh":
        if not "numDim" in kwargs:
           raise ValueError("The numDim argument is required in order to read .msh files.")
        return ReadGmsh(filename, kwargs['numDim'], integrationOrder, reducedIntegrationOrder, optimize, useMacroElements, diracPoints=points, diracTags=tags)
    else:
#        return LoadMesh(filename)
        raise ValueError("Unsupported extension .%s"%ext)




##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
# See CREDITS file for contributors and development history
#
##############################################################################

"""
Symbolic expression evaluator for escript.

This module provides the `Evaluator` class which compiles symbolic expressions
into callable functions that can be evaluated with escript Data objects.
"""

__copyright__="""Copyright (c) 2003-2026 by the esys.escript Group
https://github.com/LutzGross/esys-escript.github.io
Primary Business: Queensland, Australia"""
__license__="""Licensed under the Apache License, version 2.0
http://www.apache.org/licenses/LICENSE-2.0"""
__url__="https://github.com/LutzGross/esys-escript.github.io"
__author__="Cihan Altinay, Lutz Gross"

__all__ = ['Evaluator']

import esys.escriptcore.util as escript
import esys
from .printer import EsysEscriptPrinter
from .utils import combineData
import math

escript_functions = [ "kronecker", "identity", "zeros", "identityTensor", "identityTensor4", "unitVector", "Lsup", "sup", "inf", "testForZero", "log10", "wherePositive",
                      "whereNegative", "whereNonNegative", "whereNonPositive", "whereZero", "whereNonZero", "Abs", "erf", "sin", "cos", "tan", "asin", "acos", "atan",
                      "atan2", "sinh", "cosh", "tanh", "asinh", "acosh", "atanh", "exp", "sqrt", "log", "sign", "minval", "maxval", "length", "trace", "transpose", "swap_axes",
                      "symmetric", "nonsymmetric", "antisymmetric", "hermitian", "antihermitian", "inverse", "eigenvalues", "mult", "maximum", "minimum", "clip", "cross",
                      "inner", "outer", "matrixmult", "matrix_mult", "tensormult", "tensor_mult", "generalTensorProduct", "transposed_matrix_mult", "transposed_tensor_mult",
                      "generalTransposedTensorProduct", "matrix_transposed_mult", "tensor_transposed_mult",
                      "generalTensorTransposedProduct", "grad", "grad_n", "curl", "integrate", "interpolate", "div", "jump", "L2", "normalize", "deviatoric", "meanValue",
                      "reorderComponents", "positive", "negative", "safeDiv", "condEval", "polarToCart", "phase", "real", "imag", "conjugate" ]


translator = { n :  getattr(escript, n) for n in escript_functions }

class Evaluator(object):
    """
    Evaluates symbolic expressions with escript Data objects.

    The Evaluator compiles symbolic expressions into callable functions that
    can be evaluated efficiently with substituted values.
    """

    def __init__(self, *expressions):
        """
        Creates a symbolic evaluator.

        :param expressions: optional expressions to initialize with
        :type expressions: `Symbol` or ``sympy.Expr``
        """
        self.expressions=[]
        self.symbols=[]
        self.lambdas=[]
        self._subsdict={}
        for ex in expressions:
            self.addExpression(ex)

    def __getstate__(self):
        mydict=self.__dict__.copy()
        # lambda functions cannot be pickled
        mydict['lambdas']=[]
        return mydict

    def __setstate__(self, dict):
        expressions=dict['expressions']
        dict['expressions']=[]
        self.__dict__.update(dict)
        # regenerate lambdas
        for ex in expressions:
            self.addExpression(ex)

    def addExpression(self, expression):
        """
        Adds an expression to this evaluator.

        :param expression: symbolic expression to add
        :type expression: `Symbol` or ``sympy.Expr``
        :return: the modified Evaluator object
        :rtype: `Evaluator`
        :raise TypeError: if argument is not an expression
        """
        import sympy
        from esys import escript
        if not hasattr(expression, "atoms"):
            raise TypeError("Argument is not an expression")
        self.expressions.append(expression)
        if isinstance(expression, sympy.Function):
            sym=set()
            for arg in expression.args:
                if arg is not None:
                    sym.update(arg.atoms(sympy.Symbol))
            sym=tuple(sym)
            self.symbols.append(tuple(sym))
        else:
            sym=tuple(set(expression.atoms(sympy.Symbol)))
            self.symbols.append(sym)
        # Create modules dictionary for lambdify (required for SymPy 1.2+)
        modules = [{'combineData': combineData, 'esys': esys}, translator, math, "numpy"]
        if isinstance(expression, escript.Symbol):
            subs=expression.getDataSubstitutions()
            subs_dict={}
            for s in subs.keys():
                subs_dict[s.name] = subs[s]
            self.subs(**subs_dict)
            self.lambdas.append(sympy.lambdify(sym, expression.lambdarepr(), modules=modules, printer=EsysEscriptPrinter, dummify=False))
        else:
            self.lambdas.append(sympy.lambdify(sym, expression, modules=modules, printer=EsysEscriptPrinter, dummify=False))
        return self

    def subs(self, **args):
        """
        Performs symbol substitution.

        :keyword args: symbol name to value mappings
        :return: the modified Evaluator object
        :rtype: `Evaluator`
        """
        self._subsdict.update(args)
        return self

    def evaluate(self, evalf=False, **args):
        """
        Evaluates all expressions in this evaluator and returns the result.

        :param evalf: if True, call evalf() on Symbol results
        :type evalf: ``bool``
        :keyword args: additional symbol substitutions
        :return: the evaluated expression (single) or tuple of expressions
        :rtype: evaluated result or ``tuple``
        :raise RuntimeError: if not all symbols have values
        """
        from esys import escript
        self.subs(**args)
        res=()
        resEvaled=()
        for i in range(len(self.lambdas)):
            x=self.symbols[i]
            subslist=[self._subsdict[a.name] for a in x if a.name in self._subsdict]
            if len(x)==len(subslist):
                res+=self.lambdas[i](*subslist),
            else:
                raise RuntimeError("Not all symbols have a value")
            if evalf:
                if isinstance(res[i], escript.Symbol):
                    resEvaled+=(res[i].evalf(),)
                else:
                    resEvaled+=(res[i],)
            else:
                resEvaled+=(res[i],)
        if len(res)==1:
            return resEvaled[0]
        else:
            return resEvaled

    def __call__(self, **args):
        """
        Convenience method to substitute and evaluate at once.
        """
        return self.subs(**args).evaluate()

    def __getitem__(self, index):
        """
        Expression accessor.
        """
        return self.expressions[index]

    def __len__(self):
        """
        Returns the number of expressions in this evaluator.
        """
        return len(self.expressions)

    def __iadd__(self, expression):
        """
        Same as addExpression(expression).
        """
        return self.addExpression(expression)

    def __str__(self):
        ret="\n".join([str(e) for e in self.expressions])+"\n"
        for k in sorted(self._subsdict.keys()):
            v=self._subsdict[k]
            if v.__class__.__name__=="Data":
                ret+="%s=<Data object>"%k
            elif v.__class__.__name__=="FunctionSpace":
                ret+="%s=<FunctionSpace object>"%k
            else:
                ret+="%s=%s"%(k,v)
            ret+=", "
        return ret

#
# vim: expandtab shiftwidth=4:



##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
##############################################################################

__copyright__ = """Copyright (c) 2003-2026 by the esys.escript Group
https://github.com/LutzGross/esys-escript.github.io
Primary Business: Queensland, Australia"""
__license__ = """Licensed under the Apache License, version 2.0
http://www.apache.org/licenses/LICENSE-2.0"""
__url__ = "https://github.com/LutzGross/esys-escript.github.io"

"""

This example illustrates the use of the esys.escript.minimizers. The example
is minimization of the Rosenbrock function which is commonly used test problem
for optimization.
 
"""


from esys.escript.minimizer import MinimizerLBFGS, CostFunction
import numpy as np
from scipy.optimize import rosen, rosen_der, rosen_hess
import logging
import matplotlib
# For interactive use, you can comment out this line
matplotlib.use('agg')

mylogger = logging.getLogger('esys')
mylogger.setLevel(logging.DEBUG)


# This defines the cost function to be minimized
class RosenFunc(CostFunction):
    def __init__(self, N=4):
        super().__init__()
        self.N = N

    def getDualProduct(self,  m, g):
        """
        returns the dot product of gradient g with solution vector m:
        """
        return np.dot(g, m)

    def getArguments(self, m):
        """
        return precalculated values for current approximation m. Here nothing is returned:
        """
        return None,

    def getValue(self, m, *args):
        """
        returns the value of the cost function:
        """
        return rosen(m)

    def getGradient(self, m, *args):
        """
        returns the value of the cost function:
        """
        return rosen_der(m)

    def getInverseHessianApproximation(self, r, m, *args, initializeHessian=False):
        """
        returns a vectore of an approximation of the Hessian inverse at m for vector r.
        """
        # here the Hessian is updated when at the first iteration step and when
        # the BFGS is restarted.
        if initializeHessian:
            self.invH = np.linalg.inv(rosen_hess(m))
        return self.invH.dot(r)

    def getNorm(self, m):
        """
        return the solution of m
        """
        return np.linalg.norm(m, np.inf)

# ... callback function is used to collect the table of iteration count vs. value of the cost function
TABLE = []
def myCallback(iterCount, m, dm, Fm, grad_Fm, norm_m, args_m, failed):
    if not failed:
        TABLE.append( [ iterCount, Fm ] )

#
# Create an instance of the cost function
#
F = RosenFunc(N=20)

#... Create an instance of the solver
#
solve = MinimizerLBFGS(F, iterMax=300, logger=mylogger)
solve.setTolerance(m_tol=1e-7)
solve.setCallback(myCallback)
#... run the solver:
m0 = np.zeros((F.N,))
solve.run(m0)
m = solve.getResult()
print("m = ", m)
print("close to true solution? ", np.allclose(m, np.ones( (F.N, )) ) )

# plotting the convergency history:

import matplotlib.pyplot as plt
plt.plot([t[0] for t in TABLE], [t[1] for t in TABLE], '-')
plt.ylabel('F(m)')
plt.xlabel('iteration count')
plt.yscale('log')
plt.savefig("rosenhistory.png")



##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
##############################################################################

"""
VisIt Visualization Script for Deformation Data

This script demonstrates how to use VisIt's Python interface to create
visualizations of deformation data (stress and displacement) from VTK files.

Usage:
    visit -cli -nowin -s deformVis.py

The -nowin option prevents the visualization window from being shown, which
is not required since the purpose of the script is to save an image file.

Prerequisites:
    - VisIt must be installed
    - A deform.vtu file must exist in the current directory
"""

__copyright__ = """Copyright (c) 2003-2026 by the esys.escript Group
https://github.com/LutzGross/esys-escript.github.io
Primary Business: Queensland, Australia"""
__license__ = """Licensed under the Apache License, version 2.0
http://www.apache.org/licenses/LICENSE-2.0"""

# Open the VTU database
OpenDatabase("deform.vtu")

# Add pseudocolor plot for stress
AddPlot("Pseudocolor", "stress")
DrawPlots()

# Configure pseudocolor plot
p = PseudocolorAttributes()
p.centering = p.Nodal
SetPlotOptions(p)

# Add three-slice operator
AddOperator("ThreeSlice")
DrawPlots()

# Configure three-slice operator
t = ThreeSliceAttributes()
t.x = 0.3
t.y = 0.3
SetOperatorOptions(t)

# Add vector plot for displacement
AddPlot("Vector", "disp")
DrawPlots()

# Configure vector plot
v = VectorAttributes()
v.useStride = 1
SetPlotOptions(v)

# Configure save window attributes
s = SaveWindowAttributes()
# Change settings as required, for example:
# s.fileName = "deform_visualization"
# s.format = s.PNG
# s.width = 1024
# s.height = 768
SetSaveWindowAttributes(s)

# Save the window and exit
SaveWindow()
exit()




##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
# See CREDITS file for contributors and development history
#
##############################################################################

# This is a template configuration file for escript on Debian/GNU Linux.
# Refer to README_FIRST for usage instructions.

escript_opts_version = 203

openmp=1
paso=1
trilinos=0
umfpack=1
werror=0
verbose=0
compressed_files=0




##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
##############################################################################

"""
Multi-Physics Example: Thermo-Mechanical Coupling

This example demonstrates time-dependent coupling of thermal and mechanical
physics using DataCoupler for communication between domains. The thermal
solver computes temperature evolution while the mechanical solver calculates
thermal stresses. The mechanical model runs one time-step behind to use the
temperature from the previous step (staggered coupling).

Usage:
    # Standard mode (only rank 0 output visible):
    ./bin/run-escript -n 8 thermo_mechanical.py

    # Verbose mode (see output from all domains):
    ./bin/run-escript -n 8 -a thermo_mechanical.py

This uses 4 processes for thermal domain and 4 for mechanical domain.
"""

__copyright__ = """Copyright (c) 2003-2026 by the esys.escript Group
https://github.com/LutzGross/esys-escript.github.io
Primary Business: Queensland, Australia"""
__license__ = """Licensed under the Apache License, version 2.0
http://www.apache.org/licenses/LICENSE-2.0"""

from mpi4py import MPI
from esys.escript import *
from esys.escript import MPIDomainArray, DataCoupler
from esys.escript.linearPDEs import LinearPDE
from esys.escript.util import identityTensor4
from esys.ripley import Rectangle
from esys.weipa import saveVTK

world_comm = MPI.COMM_WORLD
world_rank = world_comm.Get_rank()
world_size = world_comm.Get_size()

# Create output directory
mkDir("output")

# Create domain array: 2 domains (thermal and mechanical)
# Domain 0: thermal solver
# Domain 1: mechanical solver
num_domains = 2
domain_array = MPIDomainArray(numDomains=num_domains, comm=world_comm)

my_domain_idx = domain_array.getDomainIndex()
domain_comm = domain_array.getDomainComm()
subdomain_comm = domain_array.getSubdomainComm()
domain_rank = domain_comm.Get_rank()

is_thermal = (my_domain_idx == 0)

# Print initialization from each domain's rank 0
if domain_rank == 0:
    print(f"Domain {my_domain_idx} ({'thermal' if is_thermal else 'mechanical'}) "
          f"initialized with {domain_comm.Get_size()} processes", flush=True)

# Create DataCoupler for inter-domain communication
coupler = DataCoupler(domain_array)

# Create domain on domain-specific communicator
domain = Rectangle(n0=80, n1=80, l0=1.0, l1=1.0, comm=domain_comm)
x = domain.getX()

# Time stepping parameters
dt = 0.00025      # Time step
n_steps = 200  # Number of time steps

if is_thermal:
    # THERMAL SOLVER: Transient heat equation
    # du/dt = kappa * nabla^2(u) + Q

    pde = LinearPDE(domain)
    kappa = 1.0  # Thermal diffusivity
    pde.setValue(A=kappa*kronecker(2), D=1/dt)

    # Initial temperature: hot spot in center
    T = 20.0 + 100.0*exp(-50*((x[0]-0.5)**2 + (x[1]-0.5)**2))
    T_old = T

    for step in range(n_steps):
        # Heat source (decaying with time)
        Q = 10.0 * exp(-0.1*step*dt) * \
            exp(-50*((x[0]-0.5)**2 + (x[1]-0.5)**2))

        # Backward Euler: (T - T_old)/dt = kappa*nabla^2(T) + Q
        pde.setValue(Y=T_old/dt + Q)
        T = pde.getSolution()

        # Send temperature field to mechanical solver using DataCoupler
        # Interpolate to Function space for consistent transfer
        T_data = interpolate(T, Function(domain))
        coupler.send(T_data, dest_domain_index=1, tag=step)

        T_max = Lsup(T)
        if domain_rank == 0 and step % 10 == 0:
            print(f"Thermal step {step}: T_max = {T_max:.2f}")
        T_old = T

else:
    # MECHANICAL SOLVER: Thermal stress (one step behind)
    # Solve: -div(sigma) = 0, sigma = E*epsilon - alpha*E*Delta_T

    pde = LinearPDE(domain, numEquations=2)
    E = 70e9        # Young's modulus (Pa)
    nu = 0.3        # Poisson's ratio
    alpha = 23e-6   # Thermal expansion coefficient (1/K)
    T0 = 20.0       # Reference temperature

    # Elastic stiffness tensor (plane stress)
    # C_ijkl = λ δ_ij δ_kl + μ (δ_ik δ_jl + δ_il δ_jk)
    lam = E*nu/(1-nu**2)       # Lamé's first parameter (plane stress)
    mu = E/(2*(1+nu))           # Shear modulus

    # Construct rank-4 elastic stiffness tensor
    d = 2  # 2D
    C = lam * kronecker(d).reshape(d,1,d,1) * kronecker(d).reshape(1,d,1,d) + \
        mu * (identityTensor4(d) + identityTensor4(d).transpose(0,1,3,2))

    pde.setValue(A=C)
    # Fixed displacement on left edge
    x = domain.getX()
    pde.setValue(q=whereZero(x[0])*[1,1], r=[0.0, 0.0])
    for step in range(0, n_steps):
        # Receive temperature from thermal solver (previous step) using DataCoupler
        # DataCoupler handles the receive and distribution across all ranks
        T_data = coupler.receive(Function(domain), source_domain_index=0, tag=step)
        # Thermal stress (prestress due to thermal expansion)
        Delta_T = T_data - T0
        sigma_thermal = (lam + 2./3.*mu) * alpha * Delta_T * kronecker(2)

        # Set thermal stress as prestress (X coefficient)
        pde.setValue(X=sigma_thermal)

        # Solve for displacement
        u = pde.getSolution()
        u_max = Lsup(length(u))
        if domain_rank == 0 and step % 10 == 0:
            print(f"Mechanical step {step}: u_max = {u_max:.2e} m")

        # Save displacement and temperature fields
        if step % 10 == 0:
            saveVTK(f"output/thermal_{step//10:04d}.vtu", u=u, T=T_data)

world_comm.Barrier()

if world_rank == 0:
    print("Thermo-mechanical coupling completed successfully")




##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
# See CREDITS file for contributors and development history
#
##############################################################################


"""
Symbolic mathematical constants for escript.

This module provides symbolic versions of common mathematical constants
that can be used in symbolic expressions.

:var pi: symbolic representation of pi (3.14159...)
:var e: symbolic representation of Euler's number e (2.71828...)
"""

from .symbol import Symbol
import sympy
pi=Symbol(sympy.pi)
e=Symbol(sympy.E)




##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
# See CREDITS file for contributors and development history
#
##############################################################################

# This is a template configuration file for escript on OpenSUSE.
# Refer to installation.md for usage instructions.
#
# Prerequisites:
#   sudo zypper in python3-devel python3-numpy python3-scipy python3-sympy python3-matplotlib
#   sudo zypper in gcc gcc-c++ gcc-fortran scons cmake
#   sudo zypper in libboost_python3-devel libboost_numpy3-devel libboost_random-devel
#   sudo zypper in hdf5-devel netcdf-devel suitesparse-devel lapack-devel zlib-devel metis-devel
#   sudo zypper in openmpi-devel python3-mpi4py
#
# For no-MPI build, use opensuse_nompi_options.py instead.

import subprocess

escript_opts_version = 203

# Python configuration - auto-detect version
pythoncmd = '/usr/bin/python3'
p = subprocess.run([pythoncmd, '-V'], capture_output=True, text=True)
subversion = p.stdout.split(' ')[1].split('.')[1]
print(f"Python subversion = {subversion}")

pythonlibpath = ['/usr/lib64']
pythonlibname = [f'python3.{subversion}']
pythonincpath = [f'/usr/include/python3.{subversion}']

# Compiler settings
openmp = True
werror = 0

# Boost configuration - OpenSUSE uses different library naming
boost_libs = ['boost_python-py3']
boost_prefix = ['/usr/include', '/usr/lib64']

# MPI configuration
mpi = 'OPENMPI'
mpi_prefix = ['/usr/lib64/mpi/gcc/openmpi4/include', '/usr/lib64/mpi/gcc/openmpi4/lib64']
mpi_libs = ['mpi_cxx', 'mpi']
mpi4py = True

# Solver configuration
paso = True
build_trilinos = 'check'

# UMFPACK direct solver
umfpack = True
umfpack_prefix = ['/usr/include/suitesparse', '/usr/lib64']
umfpack_libs = ['umfpack', 'amd']

# MUMPS sequential solver
mumps_seq = True
mumps_seq_libs = ['mumps_common', 'cmumps_seq', 'dmumps_seq', 'zmumps_seq', 'pord']

# NetCDF configuration
netcdf = True
netcdf_prefix = ['/usr/include', '/usr/lib64']
netcdf_libs = ['netcdf_c++4', 'netcdf']

# HDF5 configuration
hdf5 = True
hdf5_prefix = ['/usr/include', '/usr/lib64']
hdf5_libs = ['hdf5_cpp', 'hdf5']

# LAPACK configuration
lapack = 'auto'

# METIS configuration
metis = True
metis_prefix = ['/usr/include', '/usr/lib64']
metis_libs = ['metis']

# zlib configuration
zlib = True
zlib_prefix = ['/usr/include', '/usr/lib64']
zlib_libs = ['z']

# Optional features
sympy = True

# Clean up temporary variables (SCons treats all module-level variables as options)
del p, subversion



##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
##############################################################################

"""
Master-Slave MPI Pattern for Magnetotelluric (MT) Problem

This example demonstrates how to use mpi4py with escript to implement a
master-slave workflow. The master process distributes frequency values to
slave processes, which each solve an MT problem independently. Results are
collected by the master and plotted together.

Usage:
    mpirun -n 11 python3 MT_master_slave.py

This runs with 1 master + 10 slaves for 10 different frequencies.
"""

__copyright__ = """Copyright (c) 2003-2026 by the esys.escript Group
https://github.com/LutzGross/esys-escript.github.io
Primary Business: Queensland, Australia"""
__license__ = """Licensed under the Apache License, version 2.0
http://www.apache.org/licenses/LICENSE-2.0"""

from mpi4py import MPI
from esys.escript import *
from esys.finley import Rectangle
from esys.escript.pdetools import Locator
from esys.escript.linearPDEs import LinearSinglePDE
import numpy as np
import matplotlib.pyplot as plt

# MPI Setup
comm_world = MPI.COMM_WORLD
rank = comm_world.Get_rank()
size = comm_world.Get_size()

# Constants for MT problem
L0 = 80000   # horizontal extent [m]
L1 = 40000   # depth [m]
NE0 = 400    # number of elements in horizontal direction
NE1 = 200    # number of elements in vertical direction

rho_b = 100  # background resistivity [Ohm m]
rho_a = 0.5  # resistivity in the anomaly [Ohm m]
D = 250.     # depth of the top edge of the anomaly [m]
W = 1000.    # width of the anomaly [m]
H = 2000.    # vertical extent of the anomaly [m]
OFFSET = 1000  # offset of the central vertical axis of the anomaly

Mu0 = 4 * np.pi * 1e-7
h0, h1 = L0/NE0, L1/NE1


def solve_MT_problem(frequency):
    """
    Solve the MT problem for a given frequency on MPI.COMM_SELF.

    This function creates its own escript domain using MPI.COMM_SELF,
    ensuring that each slave process runs independently.

    Parameters
    ----------
    frequency : float
        Frequency in Hz

    Returns
    -------
    tuple
        (x_positions, apparent_resistivity, phase)
    """
    # Create domain using COMM_SELF (single process domain)
    domain = Rectangle(n0=NE0, n1=NE1, l0=L0, l1=L1, comm=MPI.COMM_SELF)

    # Set up PDE (complex-valued)
    pde = LinearSinglePDE(domain, isComplex=True)
    pde.setValue(D=1j*2*np.pi*frequency*Mu0)

    # Define anomaly mask
    X = ReducedFunction(domain).getX()
    m1 = whereNonPositive(X[1] - (L1-D))
    m2 = whereNonNegative(X[1] - (L1-D-H))
    m3 = whereNonNegative(X[0] - (L0/2+OFFSET-W/2))
    m4 = whereNonPositive(X[0] - (L0/2+OFFSET+W/2))
    m = m1 * m2 * m3 * m4

    # Set resistivity distribution
    rho = rho_b * (1-m) + rho_a * m
    rho = interpolate(rho, Function(domain))
    pde.setValue(A=rho*np.eye(2))

    # Set boundary condition: magnetic field = 1 at top
    x = domain.getX()
    mD = whereZero(x[1] - L1)
    pde.setValue(q=mD, r=1)

    # Solve for magnetic field Hx
    Hx = pde.getSolution()

    # Calculate electric field and impedance
    Ey = rho * grad(Hx, ReducedFunction(domain))[1]
    Zyx = Ey / Hx

    # Extract transect along top of domain
    locations_in_transect = [(h0*k + h0/2., L1) for k in range(0, NE0, 2)]
    locator_transect = Locator(where=ReducedFunction(domain),
                               x=locations_in_transect)

    x_ts = np.array([x[0] for x in locator_transect.getX()])
    Zyx_ts = np.array(locator_transect(Zyx))

    # Calculate apparent resistivity and phase
    rho_a_ts = 1./(2*np.pi*frequency*Mu0) * abs(Zyx_ts)**2
    phi_ts = np.angle(Zyx_ts, deg=True)

    return x_ts, rho_a_ts, phi_ts


def master_process():
    """
    Master process: distribute frequencies to slaves and collect results.
    """
    # Define frequencies to compute (logarithmic spacing)
    frequencies = np.logspace(-1, 1, 10)  # 0.1 Hz to 10 Hz
    n_frequencies = len(frequencies)

    # Check we have enough processes
    n_slaves = size - 1
    if n_slaves < 1:
        print("Error: Need at least 2 MPI processes (1 master + 1 slave)")
        print(f"Running with {size} process(es)")
        comm_world.Abort(1)

    print(f"Master: Running MT problem for {n_frequencies} frequencies")
    print(f"Master: Using {n_slaves} slave process(es)")
    print(f"Frequencies: {frequencies}")

    # Storage for results
    results = []

    # Send frequencies to slaves (round-robin distribution)
    tasks_sent = 0
    tasks_received = 0

    # Initially send one task to each slave
    for slave_rank in range(1, min(n_slaves + 1, n_frequencies + 1)):
        if tasks_sent < n_frequencies:
            freq = frequencies[tasks_sent]
            print(f"Master: Sending f={freq:.2f} Hz to slave {slave_rank}")
            comm_world.send(freq, dest=slave_rank, tag=1)
            tasks_sent += 1

    # Receive results and send new tasks
    while tasks_received < n_frequencies:
        # Receive result from any slave
        status = MPI.Status()
        result = comm_world.recv(source=MPI.ANY_SOURCE, tag=2, status=status)
        slave_rank = status.Get_source()
        tasks_received += 1

        freq, x_ts, rho_a_ts, phi_ts = result
        results.append((freq, x_ts, rho_a_ts, phi_ts))
        print(f"Master: Received result for f={freq:.2f} Hz from slave {slave_rank} "
              f"({tasks_received}/{n_frequencies})")

        # Send next task to this slave if available
        if tasks_sent < n_frequencies:
            freq = frequencies[tasks_sent]
            print(f"Master: Sending f={freq:.2f} Hz to slave {slave_rank}")
            comm_world.send(freq, dest=slave_rank, tag=1)
            tasks_sent += 1

    # Send termination signal to all slaves
    for slave_rank in range(1, size):
        comm_world.send(None, dest=slave_rank, tag=1)

    print("Master: All results collected")

    # Sort results by frequency
    results.sort(key=lambda x: x[0])

    # Plot all transects together
    plot_results(results)


def slave_process():
    """
    Slave process: receive frequencies, solve MT problem, send results back.
    """
    while True:
        # Receive frequency from master
        freq = comm_world.recv(source=0, tag=1)

        # Check for termination signal
        if freq is None:
            print(f"Slave {rank}: Received termination signal")
            break

        print(f"Slave {rank}: Solving MT problem for f={freq:.2f} Hz")

        # Solve MT problem
        x_ts, rho_a_ts, phi_ts = solve_MT_problem(freq)

        print(f"Slave {rank}: Finished f={freq:.2f} Hz, sending results")

        # Send results back to master
        result = (freq, x_ts, rho_a_ts, phi_ts)
        comm_world.send(result, dest=0, tag=2)


def plot_results(results):
    """
    Plot apparent resistivity and phase for all frequencies.

    Parameters
    ----------
    results : list of tuples
        Each tuple contains (freq, x_ts, rho_a_ts, phi_ts)
    """
    import matplotlib
    matplotlib.use('Agg')  # Non-interactive backend

    # Create figure with two subplots
    fig, (ax1, ax2) = plt.subplots(2, 1, figsize=(12, 10))

    # Color map for different frequencies
    colors = plt.cm.viridis(np.linspace(0, 1, len(results)))

    # Plot apparent resistivity
    for (freq, x_ts, rho_a_ts, phi_ts), color in zip(results, colors):
        ax1.plot(x_ts/1000, rho_a_ts, label=f'{freq:.2f} Hz',
                color=color, linewidth=2)

    ax1.set_xlabel('Distance [km]', fontsize=12)
    ax1.set_ylabel('Apparent Resistivity [Ω·m]', fontsize=12)
    ax1.set_title('MT Apparent Resistivity Transects', fontsize=14, fontweight='bold')
    ax1.legend(loc='best', fontsize=10, ncol=2)
    ax1.grid(True, alpha=0.3)
    ax1.set_yscale('log')

    # Plot phase
    for (freq, x_ts, rho_a_ts, phi_ts), color in zip(results, colors):
        ax2.plot(x_ts/1000, phi_ts, label=f'{freq:.2f} Hz',
                color=color, linewidth=2)

    ax2.set_xlabel('Distance [km]', fontsize=12)
    ax2.set_ylabel('Phase [degrees]', fontsize=12)
    ax2.set_title('MT Phase Transects', fontsize=14, fontweight='bold')
    ax2.legend(loc='best', fontsize=10, ncol=2)
    ax2.grid(True, alpha=0.3)

    plt.tight_layout()
    plt.savefig('MT_master_slave_results.png', dpi=150)
    print("Master: Saved plot to MT_master_slave_results.png")

    # Also create a separate plot showing how anomaly signature varies with frequency
    fig2, ax = plt.subplots(figsize=(12, 6))

    # Find the approximate location of the anomaly center
    anomaly_center_km = (L0/2 + OFFSET) / 1000

    for (freq, x_ts, rho_a_ts, phi_ts), color in zip(results, colors):
        # Normalize by background value for better comparison
        rho_normalized = rho_a_ts / rho_b
        ax.plot(x_ts/1000, rho_normalized, label=f'{freq:.2f} Hz',
               color=color, linewidth=2)

    ax.axvline(x=anomaly_center_km, color='red', linestyle='--',
              linewidth=2, alpha=0.5, label='Anomaly center')
    ax.set_xlabel('Distance [km]', fontsize=12)
    ax.set_ylabel('Normalized Resistivity (ρ_a / ρ_background)', fontsize=12)
    ax.set_title('Frequency-Dependent Anomaly Response', fontsize=14, fontweight='bold')
    ax.legend(loc='best', fontsize=10, ncol=2)
    ax.grid(True, alpha=0.3)

    plt.tight_layout()
    plt.savefig('MT_master_slave_normalized.png', dpi=150)
    print("Master: Saved normalized plot to MT_master_slave_normalized.png")


# Main execution
if __name__ == '__main__':
    if rank == 0:
        # Master process
        master_process()
    else:
        # Slave process
        slave_process()

    # Synchronize all processes before exit
    comm_world.Barrier()

    if rank == 0:
        print("Master: All processes finished successfully")



##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
##############################################################################

"""
Simple Ensemble Simulation Example

This example demonstrates running multiple parameter values in parallel.
Each process group solves the same problem with a different parameter.

Usage:
    mpirun -n 8 run-escript ensemble_example.py

With 8 processes and 4 parameter values, each parameter gets 2 processes.
All 4 simulations run simultaneously.
"""

__copyright__ = """Copyright (c) 2003-2026 by the esys.escript Group
https://github.com/LutzGross/esys-escript.github.io
Primary Business: Queensland, Australia"""
__license__ = """Licensed under the Apache License, version 2.0
http://www.apache.org/licenses/LICENSE-2.0"""

from mpi4py import MPI
from esys.escript import *
from esys.escript.linearPDEs import LinearPDE
from esys.ripley import Rectangle

# Get MPI information
world_comm = MPI.COMM_WORLD
world_rank = world_comm.Get_rank()
world_size = world_comm.Get_size()

# Define parameter values to explore
parameters = [0.1, 0.5, 1.0, 2.0]
num_params = len(parameters)

# Ensure we have enough processes
if world_size < num_params:
    if world_rank == 0:
        print(f"Need at least {num_params} processes for this example")
    exit(1)

# Assign each process to a parameter
param_id = world_rank % num_params
my_parameter = parameters[param_id]

# Split communicator: processes with same parameter work together
processes_per_param = world_size // num_params
group_comm = world_comm.Split(param_id, world_rank)

# Create domain with group communicator
domain = Rectangle(n0=50, n1=50, comm=group_comm)

# Solve PDE with this parameter value
pde = LinearPDE(domain)
x = domain.getX()
pde.setValue(A=my_parameter*kronecker(2), Y=x[0]*x[1], q=whereZero(x[0]))
solution = pde.getSolution()

# Each group can now process its result independently
# Note: integrate() is collective, must be called by all processes in group
result = integrate(solution)
if not domain.isRootRank():
    result = None  # Only root rank keeps the result

# Clean up communicator
group_comm.Free()

# Gather all results to rank 0 for display
all_results = world_comm.gather((my_parameter, result), root=0)

if world_rank == 0:
    print("Ensemble simulation results:")
    for param, res in all_results:
        if res is not None:
            print(f"  Parameter {param}: Result = {res}")
    print("Ensemble simulation completed successfully")




##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
# See CREDITS file for contributors and development history
#
##############################################################################


__copyright__="""Copyright (c) 2003-2026 by the esys.escript Group
https://github.com/LutzGross/esys-escript.github.io
Primary Business: Queensland, Australia"""
__license__="""Licensed under the Apache License, version 2.0
http://www.apache.org/licenses/LICENSE-2.0"""
__url__="https://github.com/LutzGross/esys-escript.github.io"

import os, re, sys
try:
    import sysconfig
    USE_DISTUTILS = False
except ImportError:
    from distutils import sysconfig
    USE_DISTUTILS = True
from subprocess import PIPE, Popen
from SCons.Script.SConscript import Configure
from site_init import findLibWithHeader, detectModule
import subprocess

REQUIRED_BOOST = (1, 46)

def CheckComplexAcos(context):
    context.Message('Checking for working complex std::acos()... ')
    result = context.TryRun("""
#include <complex>
int main() { std::complex<double> x(0,3.14159265359), y(1.5707963,-1.8622957);
return std::abs(std::acos(x)-y) < 1e-6 ? 0:-1;}
""", '.cpp')
    # scons < 2.4 fix:
    if type(result)==tuple:
        result = result[0]
    context.Result(result)
    return result

def checkCompiler(env):
    conf = Configure(env.Clone(), custom_tests = {'CheckComplexAcos': CheckComplexAcos})
    if 'CheckCXX' in dir(conf): # exists since scons 1.1.0
        if not conf.CheckCXX():
            print("Cannot run C++ compiler '%s' (check config.log)" % (env['CXX']))
            env.Exit(1)
    else:
        if not conf.CheckFunc('printf', language='c++'):
            print("Cannot run C++ compiler '%s' (check config.log)" % (env['CXX']))
            env.Exit(1)

    conf.env['buildvars']['cxx']=conf.env['CXX']

    if conf.CheckFunc('gethostname', language='c++'):
        conf.env.Append(CPPDEFINES = ['HAVE_GETHOSTNAME'])

    if conf.CheckCXXHeader('byteswap.h'):
        checkhdr="""#include <byteswap.h>
#define SCbswap32() {int x=0;bswap_32(x);}"""
        if conf.CheckFunc('SCbswap32', header=checkhdr, language='c++'):
            conf.env.Append(CPPDEFINES = ['HAVE_BYTESWAP_H'])
    if conf.CheckCXXHeader('sys/endian.h'):
        conf.env.Append(CPPDEFINES = ['HAVE_SYS_ENDIAN_H'])
    if conf.CheckCXXHeader('libkern/OSByteOrder.h'):
        conf.env.Append(CPPDEFINES = ['HAVE_OSBYTEORDER_H'])

    if not conf.CheckComplexAcos():
        conf.env.Append(CPPDEFINES = ['ESYS_USE_BOOST_ACOS'])

    return conf.Finish()

def get_external_python_sympy(env,bin):
    if env['sympy']:
        import subprocess
        cmd=''
        cmd+='import sympy\n'
        cmd+='print(sympy.__version__)\n'
        sp=subprocess.Popen([bin, '-c', cmd], stdin=None, stderr=None, stdout=subprocess.PIPE)
        spVer=sp.stdout.readline()
        if hasattr(spVer, 'decode'):
            spVer=spVer.decode()
        import sys
        if sys.version_info[0] >= 3:
            spVer = spVer.strip()
        else:
            spVer = spVer.strip().split('.')
        quit=False
        ver1=''
        ver2=''
        count=0;
        for i in range(0,len(spVer)):
            x=spVer[i]
            if x.isdigit() is True:
                if count == 0:
                    ver1=ver1+spVer[i]
                else:
                    ver2=ver2+spVer[i]
            else:
                count=count+1
                if quit is True:
                    break
                else:
                    quit=True
                    continue
        version1=float(ver1)
        version2=float(ver2)
        if version1 == 0 and version2 < 7:
            env['sympy']=False
            env['warnings'].append("sympy version too old. Symbolic toolbox and nonlinear PDEs will not be available.")
            env.Append(CPPDEFINES = ['ESYS_NO_SYMPY'])
        #elif version1 == 1 and version2 > 2:
                                              #    env['sympy']=False
        #    env['warnings'].append("escript does not support sympy version 1.2 and higher. Found %s" % spVer)
        #    env.Append(CPPDEFINES = ['ESYS_NO_SYMPY'])
        else:
            env['sympy']=True
        return env
    else:
        return env

def call_python_config(bin=None):
    import subprocess
    cmd=''
    cmd+='import subprocess\n'
    cmd+='import os\n'
    cmd+='import sys\n'
    if USE_DISTUTILS:
        cmd+='from distutils import sysconfig\n'
    else:
        cmd += 'import sysconfig\n'
    cmd+='pyversion=sysconfig.get_python_version()\n'
    cmd+='try:\n'
    cmd+='  sp=subprocess.Popen(["python"+pyversion+"-config","--ldflags"], stdout=subprocess.PIPE)\n'
    cmd+='except:\n'
    cmd+='  pythonroot=sys.exec_prefix+"/bin/"\n'
    cmd+='  sp=subprocess.Popen([pythonroot+"python"+pyversion+"-config","--ldflags"], stdout=subprocess.PIPE)\n'
    cmd+='d=sp.stdout.readline()\n'
    cmd+='if hasattr(d, "decode"):\n'
    cmd+='    d=d.decode()\n'
    cmd+='d=d.split()\n'
    cmd+="libdirs=[z[2:] for z in d if z.startswith('-L')]\n"
    cmd+="libs=[z[2:] for z in d if z.startswith('-lpython')]\n"
    cmd+="if len(libs) == 0:\n"
    cmd+='   libs = [ "python"+pyversion]\n'
    cmd+="target=''\n"
    cmd+="libname=''\n"
    cmd+="for d in libdirs:\n"
    cmd+="  for f in libs:\n"
    cmd+="    s=os.path.join(d,'lib'+f+'.so')\n"
    cmd+="    try:\n"
    cmd+="      dummy=os.stat(s)\n"
    cmd+="      if target=='':\n"
    cmd+="        target=d\n"
    cmd+="        libname=f\n"
    cmd+="    except Exception:\n"
    cmd+="      pass\n"
    cmd+="    s=os.path.join(d,'lib'+f+'.dylib')\n"
    cmd+="    try:\n"
    cmd+="      dummy=os.stat(s)\n"
    cmd+="      if target=='':\n"
    cmd+="        target=d\n"
    cmd+="        libname=f\n"
    cmd+="    except Exception:\n"
    cmd+="      pass\n"
    if bin is None:
       exec(cmd)
       ver=str(sys.version_info[0])+'.'+str(sys.version_info[1])+'.'+str(sys.version_info[2])
       if USE_DISTUTILS:
           return (target,libname,ver, sysconfig.get_python_inc())
       else:
           return (target, libname, ver, sysconfig.get_path('include'))
    # run an external python to get its library location
    # yes we are starting another python just to run python?-config
    cmd+="if hasattr(target,'decode'):\n"
    cmd+="   target=target.decode()\n"
    cmd+="   libname=libname.decode()\n"
    cmd+="print(target)\n"
    cmd+="print(libname)\n"
    cmd+="import sys\n"
    cmd+="print(str(sys.version_info[0])+'.'+str(sys.version_info[1])+'.'+str(sys.version_info[2]))\n"
    if USE_DISTUTILS:
        cmd += "print(sysconfig.get_python_inc())\n"
    else:
        cmd += "print(sysconfig.get_path('include'))\n"
    sp=subprocess.Popen([bin, '-c', cmd], stdin=None, stderr=None, stdout=subprocess.PIPE)
    target = sp.stdout.readline()
    libname = sp.stdout.readline()
    ver = sp.stdout.readline()
    pinc = sp.stdout.readline()
    if hasattr(target, "decode"):
        target, libname, ver, pinc=target.decode().strip(), libname.decode().strip(), ver.decode().strip(), pinc.decode().strip()
    else:
        target, libname, ver, pinc=target.strip(), libname.strip(), ver.strip(), pinc.strip()
    return (target, libname, ver, pinc)

def checkPython(env):
    # First we check to see if the config file has specified
    # where to find the file. Ideally, this should be automatic
    # but we need to deal with the case where python is not in its INSTALL
    # directory.
    # Use the python scons is running
    if env['pythoncmd'] == sys.executable:
        if env['IS_WINDOWS']:
            if USE_DISTUTILS:
                python_inc_path = sysconfig.get_python_inc()
            else:
                python_inc_path = sysconfig.get_path('include')
            python_lib_path=os.path.join(sysconfig.get_config_var('prefix'), 'libs')
            python_libs=['python%s%s'%(sys.version_info[0], sys.version_info[1])]
            verstring=".".join([str(i) for i in sys.version_info[:3]])
        else:
            (python_lib_path, python_libs,verstring, python_inc_path)=call_python_config(env['pythoncmd'])

    # if we want to use a python other than the one scons is running
    # Note: we assume scons is running python 2 in the following.
    else:
        if env['IS_WINDOWS']:
            cmd = "import os, sys\n"
            if USE_DISTUTILS:
                cmd += 'from distutils import sysconfig\n'
                cmd += "print(sysconfig.get_python_inc())\n"
            else:
                # stdlib sysconfig has no get_python_inc(); the include
                # directory is available via get_path('include').
                cmd += 'import sysconfig\n'
                cmd += "print(sysconfig.get_path('include'))\n"
            cmd += "print(os.path.join(sysconfig.get_config_var('prefix'), 'libs'))\n"
            cmd += "print('python%s%s'%(sys.version_info[0], sys.version_info[1]))\n"
            cmd += "print('.'.join([str(i) for i in sys.version_info[:3]]))"
            pout = subprocess.Popen([env['pythoncmd'], '-c', cmd], stdout=subprocess.PIPE).stdout.read()
            if isinstance(pout, bytes):
                pout = pout.decode(sys.stdout.encoding)
            lines = pout.split('\n')
            python_inc_path = lines[0].strip()
            python_lib_path = lines[1].strip()
            python_libs = [lines[2].strip()]
            verstring = lines[3].strip()
        else:
            (python_lib_path, python_libs,verstring, python_inc_path)=call_python_config(env['pythoncmd'])
    if isinstance(python_inc_path, bytes):
        python_inc_path=python_inc_path.decode()
    if isinstance(python_lib_path, bytes):
        python_lib_path=python_lib_path.decode()
    env['python_version'] = verstring
    try:
        ispython3 = (verstring[0] == '3')
    except:
        ispython3 = True
    if ispython3:
        env.Append(CPPDEFINES=['ESPYTHON3'])
    env['buildvars']['python_version'] = verstring
    env['buildvars']['python'] = env['pythoncmd']
    # Check for an override from the config file.
    # Ideally, this should be automatic but we need to deal with the case
    # where python is not in its INSTALL directory
    if env['pythonlibpath'] :
        python_lib_path = env['pythonlibpath']

    if env['pythonincpath'] :
        python_inc_path = env['pythonincpath']

    if env['pythonlibname'] :
        python_libs = env['pythonlibname']

    conf = Configure(env.Clone())

    if env['sysheaderopt'] :
        conf.env.AppendUnique(CCFLAGS= [ (env['sysheaderopt'] , python_inc_path) ] )
    else:
        conf.env.AppendUnique(CPPPATH = [python_inc_path])

    conf.env.AppendUnique(LIBPATH = [python_lib_path])
    conf.env.AppendUnique(LIBS = python_libs)
    # The wrapper script needs to find the libs
    conf.env.PrependENVPath(env['LD_LIBRARY_PATH_KEY'], python_lib_path)

    # Locate numpy's include directory now so it's available for downstream
    # probes (boost_numpy ships a header that transitively pulls in
    # numpy/ndarrayobject.h, so the probe link test fails to compile unless
    # this path is already on CPPPATH).
    try:
        numpy_inc = subprocess.check_output(
            [env['pythoncmd'], '-c', 'import numpy; print(numpy.get_include())'],
            text=True).strip()
        if numpy_inc and os.path.isdir(numpy_inc):
            conf.env.AppendUnique(CPPPATH=[numpy_inc])
            env['buildvars']['numpy_inc_path'] = numpy_inc
    except Exception:
        pass

    if not conf.CheckCXXHeader('Python.h'):
        print("Cannot find python include files (tried 'Python.h' in directory %s)" % (python_inc_path))
        env.Exit(1)
    if not env['IS_WINDOWS']:
        if not conf.CheckFunc('Py_Exit', language='c++'):
            print("Cannot find python library method Py_Exit (tried %s in directory %s)" % (python_libs, python_lib_path))
            print("Warning: Py_Exit check failed but continuing (Python 3.14+ compatibility)")
            # env.Exit(1)  # Disabled: Python 3.14 has Py_Exit but CheckFunc may fail

    return conf.Finish()

def checkCudaVersion(env):
    # NVCC availability is already checked in the Tool file
    p = Popen([env['NVCC'], '-V'], stdout=PIPE)
    out,_ = p.communicate()
    env['nvcc_version'] = '(unknown version)'
    for line in out.split('\n'):
        if 'release' in line:
            version = line[line.find('release'):].strip()
            env['nvcc_version'] = version
            break
    env['buildvars']['nvcc']=env['NVCC']
    return env

def checkBoost(env):

    # Skip link check for Boost on macOS with Python 3.14+ due to environment cloning issues
    # The library exists and will link properly during actual build
    boost_inc_path, boost_lib_path = findLibWithHeader(env, env['boost_libs'], 'boost/python.hpp', env['boost_prefix'], lang='c++', try_link=False)

    if env['sysheaderopt'] == '':
        env.AppendUnique(CPPPATH = [ boost_inc_path ])
    else:
        # This is required because we can't -isystem /usr/include since it
        # breaks std includes
        if os.path.normpath(boost_inc_path) == '/usr/include':
            env.Append(CCFLAGS= [ (env['sysheaderopt'], os.path.join(boost_inc_path,'boost')) ] )
        else:
            env.Append(CCFLAGS= [ (env['sysheaderopt'], boost_inc_path ) ] )

    env.AppendUnique(LIBPATH = [ boost_lib_path ])
    env.AppendUnique(LIBS = env['boost_libs'])
    env.PrependENVPath(env['LD_LIBRARY_PATH_KEY'], boost_lib_path)

    # Try to extract the boost version from version.hpp
    boosthpp=open(os.path.join(boost_inc_path, 'boost', 'version.hpp'))
    boostversion='unknown'
    for line in boosthpp:
        ver=re.match(r'#define BOOST_VERSION (\d+)',line)
        if ver:
            boostversion=ver.group(1)
            boostversion = int(boostversion)
            maj = boostversion/100000
            minor = (boostversion/100)%1000
            sub = boostversion % 100
            env['boost_version'] = "%d.%d.%d"%(maj,minor,sub)
            if maj <= REQUIRED_BOOST[0] and minor < REQUIRED_BOOST[1]:
                print("The boost version referenced must be at least version %d.%d "%REQUIRED_BOOST + "(have %d.%d.%d)"%(maj,minor,sub))
                env.Exit(1)
    boosthpp.close()
    env['buildvars']['boost_inc_path']=boost_inc_path
    env['buildvars']['boost_lib_path']=boost_lib_path
    env['buildvars']['boostversion']=boostversion

    #=======================
    # Check for the boost numpy library
    env['have_boost_numpy']=False
    if boostversion >= 106300 and not env['disable_boost_numpy'] :
        pyv = env['python_version'].split(".")
        if env["IS_OSX"]:
            libname = f'boost_numpy{pyv[0]}{pyv[1]}'
        else:
            libname = f'boost_numpy{pyv[0]}{pyv[1]}'
        # env['boost_libs'] may be a list (from an options file) or a string
        # (when set on the scons CLI), so normalise before concatenating.
        boost_libs = env['boost_libs']
        if isinstance(boost_libs, str):
            boost_libs = [boost_libs]
        try:
            boost_numpy_inc_path, boost_numpy_lib_path = \
                findLibWithHeader(env, [ libname ] + list(boost_libs), os.path.join('boost', 'python', 'numpy.hpp'), env['boost_prefix'], lang='c++')
            env.AppendUnique(LIBS=libname)
            env.AppendUnique(boost_libs=libname)
            env.AppendUnique(CPPPATH=boost_numpy_inc_path)
            env.AppendUnique(LIBPATH=boost_numpy_lib_path)
            env.PrependENVPath(env['LD_LIBRARY_PATH_KEY'], boost_numpy_lib_path)
            env.Append(CPPDEFINES=['ESYS_HAVE_BOOST_NUMPY'])
            env['have_boost_numpy'] = True
        except Exception as e:
            print("Warning: Could not find boost/python/numpy.hpp. Building without numpy support. (reason: %s)" % e)

    if boostversion >= 107000:
        env.Append(CPPDEFINES=['ESYS_DEPRECATED_BOOST_ENDIAN'])
    if boostversion >= 107200:
        env.Append(CPPDEFINES=['ESYS_MOVED_BOOST_ENDIAN'])

    return env

def checkNumpy(env):
    if not detectModule(env, 'numpy'):
        print("Cannot import numpy. If it is installed try setting your PYTHONPATH and probably %s"%env['LD_LIBRARY_PATH_KEY'])
        env.Exit(1)

    ## check for numpy header (optional)
    # checkPython already discovered numpy's include dir and put it on
    # CPPPATH (see _add_numpy_include there), so we just verify the header
    # is reachable.
    conf = Configure(env.Clone())
    numpy_h = False
    if conf.CheckCXXHeader(['Python.h', 'numpy/ndarrayobject.h']):
        numpy_h = True
    else:
        # Last-resort: re-query numpy in case CPPPATH was clobbered.
        try:
            numpy_inc = subprocess.check_output(
                [env['pythoncmd'], '-c', 'import numpy; print(numpy.get_include())'],
                text=True).strip()
            conf.env.Append(CPPPATH=[numpy_inc])
            if conf.CheckCXXHeader(['Python.h', 'numpy/ndarrayobject.h']):
                numpy_h = True
        except Exception:
            pass
    conf.env['numpy_h'] = numpy_h
    if numpy_h:
        conf.env.Append(CPPDEFINES = ['ESYS_HAVE_NUMPY_H'])

    return conf.Finish()

def checkCUDA(env):
    try:
        cuda_inc_path,cuda_lib_path=findLibWithHeader(env, 'cudart', 'thrust/version.h', env['cuda_prefix'], lang='c++')
        env.AppendUnique(CPPPATH = [cuda_inc_path])
        env.AppendUnique(LIBPATH = [cuda_lib_path])
        env.PrependENVPath(env['LD_LIBRARY_PATH_KEY'], cuda_lib_path)
        env.Append(CPPDEFINES = ['ESYS_HAVE_CUDA'])
        env['cuda']=True
    except:
        env['cuda']=False
    return env

def checkCppUnit(env):
    try:
        cppunit_inc_path,cppunit_lib_path=findLibWithHeader(env, env['cppunit_libs'], 'cppunit/TestFixture.h', env['cppunit_prefix'], lang='c++')
        env.AppendUnique(CPPPATH = [cppunit_inc_path])
        env.AppendUnique(LIBPATH = [cppunit_lib_path])
        env.PrependENVPath(env['LD_LIBRARY_PATH_KEY'], cppunit_lib_path)
        env['cppunit']=True
    except:
        env['cppunit']=False
    return env

def checkOptionalModules(env):
    ######## scipy
    if not detectModule(env, 'scipy'):
        env['warnings'].append("Cannot import scipy.")

    ######## sympy
    if env['sympy']:
        if detectModule(env, 'sympy'):
            env['sympy'] = True
        else:
            env.Append(CPPDEFINES = ['ESYS_NO_SYMPY'])
            env['sympy']=False
            env['warnings'].append("sympy not found.")
    else:
        env['sympy']=False
        env.Append(CPPDEFINES = ['ESYS_NO_SYMPY'])
    return env

def checkForTrilinos(env):
    if not env['trilinos']:
        env['trilinos_version']='none'
        return env
    havelibs = (len(env['trilinos_libs']) > 0)
    trilinos_inc_path, trilinos_lib_path=findLibWithHeader(env,
            env['trilinos_libs'], 'Amesos2.hpp',
            env['trilinos_prefix'], lang='c++', try_link=havelibs)

    # Try to extract the trilinos version from Trilinos_version.h
    versionh = open(os.path.join(trilinos_inc_path, 'Trilinos_version.h'))
    env['trilinos_version'] = 'unknown'
    for line in versionh:
        ver = re.match(r'#define TRILINOS_MAJOR_MINOR_VERSION (\d+)', line)
        if ver:
            trilinos_version = ver.group(1)
            trilinos_version = int(trilinos_version)
            major = int(str(trilinos_version)[:2])
            minor = int(str(trilinos_version)[2:4])
            subminor = int(str(trilinos_version)[4:6])
            env['trilinos_version'] = str(major) + "." + str(minor) + "." + str(subminor)
            if major < 14:
                raise RuntimeError('Trilinos version greater than 14 expected.')
            #            if major == 14 and minor <2 :
            #                env.Append(CPPDEFINES = ['ESYS_TRILINOS_14'])
            #            elif major == 14 and minor >=2:
            #                env.Append(CPPDEFINES = ['ESYS_TRILINOS_14_2'])
            #            else:
            env.Append(CPPDEFINES=[f'ESYS_TRILINOS_{major}'])

    packages = ['Tpetra', 'Kokkos', 'Belos', 'Amesos2', 'Ifpack2', 'MueLu']

    dependencies = ['Amesos2.hpp', 'Amesos2_Solver_decl.hpp', 'BelosSolverFactory.hpp', 'BelosSolverManager.hpp', \
                    'BelosTFQMRIter.hpp', 'BelosTFQMRSolMgr.hpp', 'BelosTpetraAdapter.hpp', 'BelosTypes.hpp', \
                    'Ifpack2_Factory.hpp',  'Tpetra_KokkosCompat_DefaultNode.hpp', \
                    'MatrixMarket_Tpetra.hpp', 'MueLu_CreateTpetraPreconditioner.hpp', \
                    'Teuchos_DefaultComm.hpp', 'Teuchos_ParameterList.hpp', 'Tpetra_BlockCrsMatrix.hpp', \
                    'Teuchos_Comm.hpp', 'Teuchos_TimeMonitor.hpp', 'Tpetra_CrsMatrix_decl.hpp', \
                    'Tpetra_BlockCrsMatrix_decl.hpp',  \
                    'Tpetra_CrsGraph.hpp', 'Tpetra_CrsMatrix.hpp', 'Tpetra_RowMatrix.hpp', \
                    'TpetraExt_TripleMatrixMultiply_def.hpp', 'Tpetra_BlockVector.hpp', \
                    'Tpetra_Vector.hpp', 'Trilinos_version.h', 'Tpetra_BlockCrsMatrix_Helpers.hpp']

    env.AppendUnique(LIBPATH = [trilinos_lib_path])
    env.PrependENVPath(env['LD_LIBRARY_PATH_KEY'], trilinos_lib_path)
    env['buildvars']['trilinos_lib_path']=trilinos_lib_path


    if major > 16 or ( major == 16 and minor > 0) :
        dependencies +=[ os.path.join('kokkos', 'Kokkos_Core.hpp' ) ]
        env.AppendUnique(CPPPATH=[trilinos_inc_path, os.path.join(trilinos_inc_path, 'kokkos')])
        env['buildvars']['trilinos_inc_path'] = [trilinos_inc_path, os.path.join(trilinos_inc_path, 'kokkos')]
    else:
        env.AppendUnique(CPPPATH=[trilinos_inc_path])
        env['buildvars']['trilinos_inc_path'] = [ trilinos_inc_path ]
        dependencies +=['Kokkos_Core.hpp']

    print("Looking for the Trilinos headers...")
    for check in dependencies:
        print("Checking for %s... %s" % (check, "yes" if os.path.isfile(os.path.join(trilinos_inc_path,check)) else "no") )
        if not os.path.isfile(os.path.join(trilinos_inc_path,check)):
            print("Could not find the  Trilinos header file %s (tried looking in directory %s)" % (check, trilinos_inc_path))
            env.Exit(1)
    # Check for required Tpetra headers (standard versions required for Trilinos 16.2+)
    required_tpetra_headers = [
        'Tpetra_BlockCrsMatrix.hpp',
        'Tpetra_BlockCrsMatrix_Helpers.hpp',
        'Tpetra_BlockVector.hpp'
    ]
    for header in required_tpetra_headers:
        header_path = os.path.join(trilinos_inc_path, header)
        if os.path.isfile(header_path):
            print("Checking for %s... yes" % header)
        else:
            raise RuntimeError('Could not locate the Trilinos header %s (required for Trilinos 16.2+)' % header)

    library_list=['amesos2', 'amesos', 'anasaziepetra', 'anasazi', 'anasazitpetra', 'aztecoo', \
                  'belosepetra', 'belos', 'belostpetra', 'belosxpetra', 'epetraext', 'epetra', \
                  'galeri', 'galeri', 'ifpack2', 'ifpack2', 'ifpack', 'intrepid2', 'isorropia',\
                  'kokkosalgorithms', 'kokkoscontainers', 'kokkoscore', 'kokkoskernels', 'kokkossimd', \
                  'kokkostsqr', 'komplex', 'ml', 'ModeLaplace', 'muelu', 'muelu', 'pamgen_extras',\
                  'pamgen', 'rtop', 'sacado', 'shards', 'shylu_nodehts', 'simpi', 'stratimikosamesos2',\
                  'stratimikosamesos', 'stratimikosaztecoo', 'stratimikosbelos', 'stratimikosifpack', \
                  'stratimikosml', 'stratimikos', 'tacho', 'teko', 'teuchoscomm', 'teuchoscore', \
                  'teuchoskokkoscomm', 'teuchoskokkoscompat', 'teuchosnumerics', 'teuchosparameterlist', \
                  'teuchosparser', 'teuchosremainder', 'thyracore', 'thyraepetraext', 'thyraepetra', \
                  'thyratpetra', 'tpetraclassic', 'tpetraext', 'tpetrainout', 'tpetra', 'xpetra', 'xpetra-sup',
                  'trilinosss', 'triutils', 'xpetra', 'xpetra', 'zoltan', 'zoltan2']

    if not havelibs:
        libs = []
        for file in os.listdir(trilinos_lib_path):
            if file.endswith(".so"):
                for x in range(0, len(library_list)):
                    if file[3:-3] == library_list[x] or file[12:-3] == library_list[x]:
                        libs.append(file[3:-3])
            elif file.endswith(".dylib"):
                for x in range(0, len(library_list)):
                    if file[3:-6] == library_list[x] or file[12:-6] == library_list[x]:
                        libs.append(file[3:-6])
        env['trilinos_libs'] = libs
    # Trilinos version 14 and higher
    # if major >= 14:
    #     if not havelibs:
    #         libs = []
    #         for file in os.listdir(trilinos_lib_path):
    #             if file[-3:] == ".so":
    #                 libs.append(file[3:-3])
    #             if file[-3:] == ".dylib": # macOS
    #                 libs.append(file[3:-6])
    #         env['trilinos_libs'] = libs
    # else:
    #     if not havelibs:
    #         libs = []
    #         for pk in packages:
    #             # find out what libraries to link with...
    #             makefile = os.path.join(trilinos_inc_path, 'Makefile.export.%s'%pk)
    #             try:
    #                 for l in open(makefile, 'r').readlines():
    #                     if l.startswith("%s_LIBRARIES"%pk): # or l.startswith("Trilinos_TPL_LIBRARIES"):
    #                         lst = l.split('=')[1].strip().split()
    #                         lst = [e.replace('-l','',1) for e in lst]
    #                         libs += lst
    #                     elif l.startswith("%s_TPL_INCLUDE_DIRS"%pk):
    #                         lst = l.split('=')[1].strip().split()
    #                         lst = [e.replace('-I','',1) for e in lst]
    #                         env.AppendUnique(CPPPATH = lst)

    #             except Exception as e:
    #                 raise RuntimeError('Error reading Trilinos export Makefile\n%s'%(e))
    #         env['trilinos_libs'] = libs

    pytri_path=trilinos_lib_path+'/python'+str(sys.version_info[0])+'.'+str(sys.version_info[1])+'/site-packages/PyTrilinos'
    pytri_test_file=os.path.join(pytri_path,'Tpetra.pyc')
    if os.path.isfile(pytri_test_file):
        env.Append(PYTHONPATH=pytri_path)
    paths=sys.path
    for i in range(0,paths.__len__()):
        env.Append(PYTHONPATH=paths[i])
    env.Append(CPPDEFINES = ['ESYS_HAVE_TRILINOS'])
    # env.PrependENVPath(env['LD_LIBRARY_PATH_KEY'], trilinos_lib_path)
    # env['buildvars']['trilinos_inc_path']=trilinos_inc_path
    # env['buildvars']['trilinos_lib_path']=trilinos_lib_path
    env['buildvars']['trilinos']=int(env['trilinos'])
    return env

def checkOptionalLibraries(env):
    env['usempi'] = env['mpi']!='none'
    ######## hdf5
    hdf5_inc_path = ''
    hdf5_lib_path = ''
    if env['hdf5']:
        env.Append(CPPDEFINES=['HDF5'])
        hdf5_inc_path, hdf5_lib_path = findLibWithHeader(env, env['hdf5_libs'], 'H5Cpp.h',
                                                                 env['hdf5_prefix'], lang='c++')

        env.AppendUnique(CPPPATH=[hdf5_inc_path])
        env.AppendUnique(LIBPATH=[hdf5_lib_path])
        env.PrependENVPath(env['LD_LIBRARY_PATH_KEY'], hdf5_lib_path)
        env.Append(CPPDEFINES=['ESYS_HAVE_HDF5'])
        env['buildvars']['hdf5_inc_path'] = hdf5_inc_path
        env['buildvars']['hdf5_lib_path'] = hdf5_lib_path
    env['buildvars']['hdf5'] = int(env['hdf5'])
    ######## MKL
    mkl_inc_path=''
    mkl_lib_path=''
    if env['mkl']:
        mkl_inc_path,mkl_lib_path=findLibWithHeader(env, env['mkl_libs'], 'mkl_pardiso.h', env['mkl_prefix'], lang='c++')
        env.AppendUnique(CPPPATH = [mkl_inc_path])
        env.AppendUnique(LIBPATH = [mkl_lib_path])
        env.PrependENVPath(env['LD_LIBRARY_PATH_KEY'], mkl_lib_path)
        env.Append(CPPDEFINES = ['ESYS_HAVE_MKL'])
        env['buildvars']['mkl_inc_path']=mkl_inc_path
        env['buildvars']['mkl_lib_path']=mkl_lib_path
    env['buildvars']['mkl']=int(env['mkl'])

    ######## UMFPACK
    umfpack_inc_path=''
    umfpack_lib_path=''
    if env['umfpack']:
        umfpack_inc_path,umfpack_lib_path=findLibWithHeader(env, env['umfpack_libs'], 'umfpack.h', env['umfpack_prefix'], lang='c++')
        env.AppendUnique(CPPPATH = [umfpack_inc_path])
        env.AppendUnique(LIBPATH = [umfpack_lib_path])
        env.PrependENVPath(env['LD_LIBRARY_PATH_KEY'], umfpack_lib_path)
        env.Append(CPPDEFINES = ['ESYS_HAVE_UMFPACK'])
        env['buildvars']['umfpack_inc_path']=umfpack_inc_path
        env['buildvars']['umfpack_lib_path']=umfpack_lib_path
    env['buildvars']['umfpack']=int(env['umfpack'])

    ######## Sequential MUMPS (works with MPI builds)
    mumps_seq_inc_path=''
    mumps_seq_lib_path=''
    if env['mumps_seq']:
        # Sequential MUMPS uses mumps_seq/mpi.h
        mumps_seq_inc_path, mumps_seq_lib_path = findLibWithHeader(env, env['mumps_seq_libs'], 'mumps_seq/mpi.h', env['mumps_seq_prefix'], lang='c++')
        env.AppendUnique(CPPPATH = [mumps_seq_inc_path])
        env.AppendUnique(LIBPATH = [mumps_seq_lib_path])
        env.PrependENVPath(env['LD_LIBRARY_PATH_KEY'], mumps_seq_lib_path)
        env.Append(CPPDEFINES = ['ESYS_HAVE_MUMPS'])
        env['buildvars']['mumps_seq_inc_path']=mumps_seq_inc_path
        env['buildvars']['mumps_seq_lib_path']=mumps_seq_lib_path
    env['buildvars']['mumps_seq']=int(env['mumps_seq'])

    ######## LAPACK
    lapack_inc_path=''
    lapack_lib_path=''
    flavour = 'none'
    env['uselapack'] = False
    if env['lapack'] != 0:
        # not explicitly disabled so run the checks
        if env['longindices']:
            # you want longindices + lapack? sorry.
            if env['lapack'] == 1:
                print("LAPACK requires index type = int. Set longindices to False or disable LAPACK.")
                env.Exit(1)
        else:
            if env['mkl']:
                # we detected MKL so try the MKL header+libs
                flavour = 'mkl'
                header = 'mkl_lapack.h'
                prefix = env['mkl_prefix']
                if len(env['lapack_libs']) == 0:
                    libs = env['mkl_libs']
                else:
                    libs = env['lapack_libs']
            else:
                # try for lapacke
                flavour = 'lapacke'
                header = 'lapacke.h'
                prefix = env['lapack_prefix']
                if len(env['lapack_libs']) == 0:
                    libs = ['lapacke']
                else:
                    libs = env['lapack_libs']

            try:
                lapack_inc_path,lapack_lib_path=findLibWithHeader(env, libs, header, prefix, lang='c++')
                env['lapack_libs'] = libs
                env['uselapack'] = True
                env.AppendUnique(CPPPATH = [lapack_inc_path])
                env.AppendUnique(LIBPATH = [lapack_lib_path])
                env.Append(CPPDEFINES = ['ESYS_HAVE_LAPACK'])
                if flavour == 'mkl':
                    env.AppendUnique(CPPDEFINES = ['ESYS_MKL_LAPACK'])
                env['buildvars']['lapack_inc_path']=lapack_inc_path
                env['buildvars']['lapack_lib_path']=lapack_lib_path
            except:
                if env['lapack'] == 1:
                    raise
                # lapack was set to auto-detect so not a fatal error
                flavour = 'none'

    env['lapack'] = flavour
    env['buildvars']['lapack'] = flavour

    ######## Silo
    silo_inc_path=''
    silo_lib_path=''
    if env['silo']:
        silo_inc_path,silo_lib_path=findLibWithHeader(env, env['silo_libs'], 'silo.h', env['silo_prefix'], lang='c++')
        env.AppendUnique(CPPPATH = [silo_inc_path])
        env.AppendUnique(LIBPATH = [silo_lib_path])
        env.Append(CPPDEFINES = ['ESYS_HAVE_SILO'])
        env['buildvars']['silo_inc_path']=silo_inc_path
        env['buildvars']['silo_lib_path']=silo_lib_path
    env['buildvars']['silo']=int(env['silo'])

    ######## NetCDF
    netcdf_inc_path=''
    netcdf_lib_path=''
    if env['netcdf']:
        netcdf_inc_path,netcdf_lib_path=findLibWithHeader(env, env['netcdf_libs'], 'ncFile.h', env['netcdf_prefix'], lang='c++')
        env.AppendUnique(CPPPATH = [netcdf_inc_path])
        env.AppendUnique(LIBPATH = [netcdf_lib_path])
        env.Append(CPPDEFINES = ['ESYS_HAVE_NETCDF4'])
        env['buildvars']['netcdf_inc_path']=netcdf_inc_path
        env['buildvars']['netcdf_lib_path']=netcdf_lib_path
    env['buildvars']['netcdf']=int(bool(env['netcdf']))

    ######## VisIt
    visit_inc_path=''
    visit_lib_path=''
    if env['visit']:
        visit_inc_path,visit_lib_path=findLibWithHeader(env, env['visit_libs'], 'VisItControlInterface_V2.h', env['visit_prefix'], lang='c++')
        env.AppendUnique(CPPPATH = [visit_inc_path])
        env.AppendUnique(LIBPATH = [visit_lib_path])
        env['buildvars']['visit_inc_path']=visit_inc_path
        env['buildvars']['visit_lib_path']=visit_lib_path
    env['buildvars']['visit']=int(env['visit'])

    ######## MPI

    mpi_inc_path=''
    mpi_lib_path=''
    if env['usempi']:
        mpi_inc_path,mpi_lib_path=findLibWithHeader(env, env['mpi_libs'], 'mpi.h', env['mpi_prefix'], lang='c++')
        env.AppendUnique(CPPPATH = [mpi_inc_path])
        env.AppendUnique(LIBPATH = [mpi_lib_path])
        env.AppendUnique(LIBS = env['mpi_libs'])
        env.PrependENVPath(env['LD_LIBRARY_PATH_KEY'], mpi_lib_path)
        env.Append(CPPDEFINES = ['ESYS_MPI', 'MPI_NO_CPPBIND', 'MPICH_IGNORE_CXX_SEEK'])

        if env['mpi'] == 'OPENMPI':
            # try to get version for correct launcher arguments
            try:
                p = Popen(['orterun', '-V'], stdout=PIPE, stderr=PIPE)
                o,e = p.communicate()
                try:
                    e=e.decode()
                    ver = e.split('\n')[0].split()[-1]
                except IndexError:
                    o=o.decode()
                    ver = o.split('\n')[0].split()[-1]
                if len(ver) > 0:
                    env['orte_version'] = ver
            except OSError:
                pass

        env['buildvars']['mpi_inc_path']=mpi_inc_path
        env['buildvars']['mpi_lib_path']=mpi_lib_path
    env['buildvars']['mpi']=env['mpi']

    ######## METIS
    metis_inc_path=''
    metis_lib_path=''
    if env['metis']:
        metis_inc_path,metis_lib_path=findLibWithHeader(env, env['metis_libs'], 'metis.h', env['metis_prefix'], lang='c++')
        env.AppendUnique(CPPPATH = [metis_inc_path])
        env.AppendUnique(LIBPATH = [metis_lib_path])
        env.PrependENVPath(env['LD_LIBRARY_PATH_KEY'], metis_lib_path)

        # Try to extract the metis version from metis.h
        header=open(os.path.join(metis_inc_path, 'metis.h')).readlines()
        major,minor,sub = None,None,None
        for line in header:
            ver=re.match(r'#define METIS_VER_MAJOR\s*(\d+)',line)
            if ver:
                major = int(ver.group(1))
                continue
            ver=re.match(r'#define METIS_VER_MINOR\s*(\d+)',line)
            if ver:
                minor = int(ver.group(1))
                continue
            ver=re.match(r'#define METIS_VER_SUBMINOR\s*(\d+)',line)
            if ver:
                sub = int(ver.group(1))
                continue
        if major is not None:
            env['metis_version'] = "%d.%d.%d"%(major,minor if minor else 0,sub if sub else 0)
        else:
            env['metis_version'] = "unknown"
    else:
        env['metis_version'] = "N/A"
    env['buildvars']['metis']=int(env['metis'])
    env['buildvars']['metis_inc_path']=metis_inc_path
    env['buildvars']['metis_lib_path']=metis_lib_path

    ######## ParMETIS
    if not env['usempi']: env['parmetis'] = False
    parmetis_inc_path=''
    parmetis_lib_path=''
    if env['parmetis']:
        parmetis_inc_path,parmetis_lib_path=findLibWithHeader(env, env['parmetis_libs'], 'parmetis.h', env['parmetis_prefix'], lang='c++')
        env.AppendUnique(CPPPATH = [parmetis_inc_path])
        env.AppendUnique(LIBPATH = [parmetis_lib_path])
        env.PrependENVPath(env['LD_LIBRARY_PATH_KEY'], parmetis_lib_path)

        # Try to extract the parmetis version from parmetis.h
        header=open(os.path.join(parmetis_inc_path, 'parmetis.h')).readlines()
        major,minor,sub = None,None,None
        for line in header:
            ver=re.match(r'#define PARMETIS_MAJOR_VERSION\s*(\d+)',line)
            if ver:
                major = int(ver.group(1))
                continue
            ver=re.match(r'#define PARMETIS_MINOR_VERSION\s*(\d+)',line)
            if ver:
                minor = int(ver.group(1))
                continue
            ver=re.match(r'#define PARMETIS_SUBMINOR_VERSION\s*(\d+)',line)
            if ver:
                sub = int(ver.group(1))
                continue
        if major is not None:
            env['parmetis_version'] = "%d.%d.%d"%(major,minor,0 if sub is None else sub)
            if env['longindices']:
                # ParMETIS version 3.x does not support 64-bit indices
                if major < 4:
                    print("Sorry, cannot use ParMETIS version < 4.0 with 64-bit index types. Set longindices to False or disable ParMETIS.")
                    env.Exit(1)
                else:
                    # check if ParMETIS was built with 64-bit indices
                    conf = Configure(env.Clone())
                    idxsize=conf.CheckTypeSize('idx_t', '#include <parmetis.h>', 'C++')
                    if idxsize != 8:
                        print("Sorry, ParMETIS was not compiled with 64-bit indices. Set longindices to False or disable/rebuild ParMETIS.")
                        env.Exit(1)
        else:
            env['parmetis_version'] = "unknown"

        env.Append(CPPDEFINES = ['ESYS_HAVE_PARMETIS'])
        env['buildvars']['parmetis_inc_path']=parmetis_inc_path
        env['buildvars']['parmetis_lib_path']=parmetis_lib_path
    env['buildvars']['parmetis']=int(env['parmetis'])

    ######## Scotch/PT-Scotch
    if not env['usempi']: env['scotch'] = False
    scotch_inc_path=''
    scotch_lib_path=''
    if env['scotch']:
        scotch_inc_path,scotch_lib_path=findLibWithHeader(env, env['scotch_libs'], 'scotch.h', env['scotch_prefix'], lang='c++')
        env.AppendUnique(CPPPATH = [scotch_inc_path])
        env.AppendUnique(LIBPATH = [scotch_lib_path])
        env.PrependENVPath(env['LD_LIBRARY_PATH_KEY'], scotch_lib_path)

        # Try to extract the scotch version from scotch.h
        header=open(os.path.join(scotch_inc_path, 'scotch.h')).readlines()
        major,minor,patch = None,None,None
        for line in header:
            ver=re.match(r'#define SCOTCH_VERSION\s+(\d+)',line)
            if ver:
                major = int(ver.group(1))
                continue
            ver=re.match(r'#define SCOTCH_RELEASE\s+(\d+)',line)
            if ver:
                minor = int(ver.group(1))
                continue
            ver=re.match(r'#define SCOTCH_PATCHLEVEL\s+(\d+)',line)
            if ver:
                patch = int(ver.group(1))
                continue
        if major is not None:
            env['scotch_version'] = "%d.%d.%d"%(major,minor if minor else 0,patch if patch else 0)
        else:
            env['scotch_version'] = "unknown"

        env.Append(CPPDEFINES = ['ESYS_HAVE_SCOTCH'])
        env['buildvars']['scotch_inc_path']=scotch_inc_path
        env['buildvars']['scotch_lib_path']=scotch_lib_path
    else:
        env['scotch_version'] = "N/A"
    env['buildvars']['scotch']=int(env['scotch'])

    ######## zlib
    if env['zlib']:
        try:
            zlib_inc_path, zlib_lib_path = findLibWithHeader(env, env['zlib_libs'], 'zlib.h', env['zlib_prefix'], lang='c')
            env.Append(CPPDEFINES = ['ESYS_HAVE_ZLIB'])
            env.AppendUnique(CPPPATH=[zlib_inc_path])
            env.AppendUnique(LIBPATH=[zlib_lib_path])
            env.AppendUnique(LIBS=env['zlib_libs'])
            env['buildvars']['zlib_inc_path'] = zlib_inc_path
            env['buildvars']['zlib_lib_path'] = zlib_lib_path
        except RuntimeError as e:
            env['zlib'] = False
    env['buildvars']['zlib']=int(env['zlib'])
    env['buildvars']['sympy']=int(env['sympy'])
    ######## boost::iostreams
    if env['compressed_files']:
        try:
            boost_inc_path, boost_lib_path = findLibWithHeader(env, env['compression_libs'], 'boost/iostreams/filter/gzip.hpp', env['boost_prefix'], lang='c++')
            env.Append(CPPDEFINES = ['ESYS_HAVE_BOOST_IO'])
        except RuntimeError as e:
            env['compressed_files'] = False
    env['buildvars']['compressed_files']=int(env['compressed_files'])

    ######## Trilinos
    env = checkForTrilinos(env)

    ######## mpi4py
    if env['mpi4py']:
        try:
            import mpi4py
            mpi4py_inc_path=mpi4py.get_include()
            env.AppendUnique(CPPPATH = [mpi4py_inc_path])
            env.Append(CPPDEFINES = ['ESYS_HAVE_MPI4PY'])
            env['mpi4py'] = True
        except RuntimeError as e:
            env['warnings'].append("Could not import mpi4py.")

    return env

def checkPDFLatex(env):
    if 'PDF' in dir(env) and '.tex' in env.PDF.builder.src_suffixes(env):
        env['pdflatex']=True
    else:
        env['pdflatex']=False
    return env




##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
# See CREDITS file for contributors and development history
#
##############################################################################

"""
Symbolic mathematics support for escript.

This package provides symbolic computation capabilities for escript using
sympy as the backend. It allows defining PDEs symbolically and automatic
computation of Jacobians for nonlinear PDE solvers.
"""


__copyright__="""Copyright (c) 2003-2026 by the esys.escript Group
https://github.com/LutzGross/esys-escript.github.io
Primary Business: Queensland, Australia"""
__license__="""Licensed under the Apache License, version 2.0
http://www.apache.org/licenses/LICENSE-2.0"""
__url__="https://github.com/LutzGross/esys-escript.github.io"

from .symbol import *
from . import functions as symfn
from . import symconstants as Symconsts
from .pretty import pretty_print, pprint
from .utils import *
from .evaluator import *

#from .evaluator import *

#__nodocorecursion=['symbol', 'evaluator']
#
#from esys.escriptcore.start import HAVE_SYMBOLS
#if HAVE_SYMBOLS:
    #from . import functions as symfn
    #from . import symconstants as Symconsts
    #from .pretty import pretty_print, pprint
    #from .utils import *

    #prefer escript's implementation of functions such as 'sign' etc.
    #from sympy.utilities.lambdify import MODULES
    #ESCRIPT_NAMESPACE = {}
    #ESCRIPT_DEFAULT = {}
    #ESCRIPT_TRANSLATIONS = {
        #"ln":"log",
    #}

    #if len(MODULES['math'])==3:
        #MODULES['escript']=(ESCRIPT_NAMESPACE, ESCRIPT_TRANSLATIONS,('from esys.escript import *',))
    #else:
        #MODULES['escript']=(ESCRIPT_NAMESPACE, ESCRIPT_DEFAULT, ESCRIPT_TRANSLATIONS,('from esys.escript import *',))

    #del ESCRIPT_NAMESPACE
    #del ESCRIPT_DEFAULT
    #del ESCRIPT_TRANSLATIONS
    #del MODULES


#
# vim: expandtab shiftwidth=4:




##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
# See CREDITS file for contributors and development history
#
##############################################################################

"""
Startup configuration and environment setup for escript core packages.

This module performs initialization tasks required before other escript
packages are loaded, including setting library load flags for Intel MKL
compatibility and optional memory profiling setup.

:note: No escript packages should be imported in this module to prevent
       circular import issues.
"""


__copyright__="""Copyright (c) 2003-2026 by the esys.escript Group
https://github.com/LutzGross/esys-escript.github.io
Primary Business: Queensland, Australia"""
__license__="""Licensed under the Apache License, version 2.0
http://www.apache.org/licenses/LICENSE-2.0"""
__url__="https://github.com/LutzGross/esys-escript.github.io"

# This file includes setup and tweaks that are required for escript core packages
# No escript packages should be imported here

from esys.escriptcore.escriptcpp import hasFeature
try:
    # this is required so newer intel MKL libraries find necessary symbols
    import ctypes, sys
    sys.setdlopenflags(sys.getdlopenflags()|ctypes.RTLD_GLOBAL)
except:
    pass

if hasFeature('sympy'):

    try:
        import sympy as sp
        HAVE_SYMBOLS=True
    except ImportError:
        HAVE_SYMBOLS=False
else:
    HAVE_SYMBOLS=False

# To have this function called automatically
def escriptLogMemoryStatusNow(prefix='memescript'):
    """
    Logs the current memory status of the escript process to a file.

    On POSIX systems, this reads /proc/<pid>/status and writes it to a file
    named <prefix>.<pid>. This is useful for memory profiling and debugging.

    :param prefix: prefix for the output filename
    :type prefix: ``str``
    """
    import os
    if os.name=='posix':
        pid=os.getpid()
        os.system("cat /proc/%d/status > %s.%d"%(pid,prefix,pid))
    
try:
    import os
    if 'escriptExitProfiling' in os.environ:
        import atexit
        atexit.register(escriptLogMemoryStatusNow)
except:
    pass
  





##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
# See CREDITS file for contributors and development history
#
##############################################################################


__copyright__="""Copyright (c) 2003-2026 by the esys.escript Group
https://github.com/LutzGross/esys-escript.github.io
Primary Business: Queensland, Australia"""
__license__="""Licensed under the Apache License, version 2.0
http://www.apache.org/licenses/LICENSE-2.0"""
__url__="https://github.com/LutzGross/esys-escript.github.io"

import subprocess

"""
this script generates a variety of meshes in the fly format on a unit square with a 
fixed number of elements NE (approximatively).

This script will not run under MPI.

generated file names are:

   mesh_D<d>_o<O>_T<t>_Contacts<c>_Rich<r>.fly

wehere <d>= spatial dimension 2,3
       <o>= order 1,2, 2F
       <t>= element type = hex, tet
       <c>= Yes if contact elements are present, No otherwise.
       <r>= YEs is rich surface elements are used, No otherwise

any item with x[d]<CUT will be tagged with 1 the rest will be marked 100.
"""

MESH_DIRECTORY="./tmp_meshes"
NE=10
CUT=0.5

from esys.escript import *
from esys.finley import Rectangle, Brick, Merge, JoinFaces
from esys.finley import ReadGmsh
from esys.weipa import saveVTK
import os
GEO = { 2 :
r"""
Mesh.MshFileVersion = 2.2;
Mesh.ElementOrder = {o};
l_X={l_X};
element_size = {element_size};
Point(10)=\{0.,0.,0.,element_size\};
Point(11)=\{l_X,0.,0.,element_size\};
Point(12)=\{l_X,1.,0.,element_size\};
Point(13)=\{0.,1.,0.,element_size\};
Line(1)=\{10,11\};
Line(12)=\{11,12\};
Line(23)=\{12,13\};
Line(30)=\{13,10\};
CurveLoop(10) = \{1,12,23,30\};
Plane_Surface(1) = \{10\};
Physical Curve(100) = \{1, 12, 23, 30\};
Physical Surface(200) = \{1\};
""",
        3 :
r"""
Mesh.MshFileVersion = 2.2;
Mesh.ElementOrder = {o};
l_X={l_X};
element_size = {element_size};
Point(1000)=\{ 0.,0.,0.,element_size\};
Point(1100)=\{l_X,0.,0.,element_size\};
Point(1010)=\{0.,1.,0.,element_size\};
Point(1110)=\{l_X,1.,0.,element_size\};
Point(1001)=\{0.,0.,1.,element_size\};
Point(1101)=\{l_X,0.,1.,element_size\};
Point(1011)=\{0.,1.,1.,element_size\};
Point(1111)=\{l_X,1.,1.,element_size\};

Line(10)=\{1000,1100\};
Line(20)=\{1100,1110\};
Line(30)=\{1110,1010\};
Line(40)=\{1010,1000\};
Line(11)=\{1000,1001\};
Line(21)=\{1100,1101\};
Line(31)=\{1110,1111\};
Line(41)=\{1010,1011\};
Line(12)=\{1001,1101\};
Line(22)=\{1101,1111\};
Line(32)=\{1111,1011\};
Line(42)=\{1011,1001\};
CurveLoop(10)=\{10,20,30,40\};
CurveLoop(11)=\{12,22,32,42\};
CurveLoop(12)=\{10,21,-12,-11\}; 
CurveLoop(13)=\{30,41,-32,-31\};
CurveLoop(14)=\{11,-42,-41,40\};
CurveLoop(15)=\{-21,20,31,-22\};

PlaneSurface(1)=\{-10\};
PlaneSurface(2)=\{11\};
PlaneSurface(3)=\{12}\;
PlaneSurface(4)=\{13}\;
PlaneSurface(5)=\{14\};
PlaneSurface(6)=\{15\};
SurfaceLoop(10) = \{1,2,3,4,5,6\}; 
Volume(1) = \{10\};
Physical Surface(200) = \{1,2,3,4,5,6\}; 
Physical Volume(300) = \{1\};
"""
        }
def getMesh(NE_X, NE_Y, t,d,o,fullOrder,r,l_X):
   if t == "Hex":
       if d == 2:
         dom=Rectangle(n0=NE_X, n1=NE_Y, l0=l_X,order=o, useFullElementOrder=fullOrder,useElementsOnFace=r,optimize=True)
       else:
         Brick()
         dom=Brick(n0=NE_X, n1=NE_Y, n2=NE_Y,l0=l_X,order=o, useFullElementOrder=fullOrder,useElementsOnFace=r,optimize=True)
   else:
      geospec = GEO[d].format( {"o": o, "l_X" : l_X,  "element_size" : min(l_X/max(3,NE_X),1./max(3,NE_Y)) } )
      FF=open("tester.geo", 'w')
      FF.write(geospec)
      FF.close()
      import subprocess
      rp = subprocess.run(["gmsh", f"-{o}", "-optimize_netgen", "-o", "tester.msh", FF.name], stdout=subprocess.PIPE)
      dom = ReadGmsh("tester.msh", d, optimize=True)
   return dom
    
# test if out put dir exists:
if not os.path.isdir(MESH_DIRECTORY): os.mkdir(MESH_DIRECTORY)

for d in [2,3]:
   for  o in ["1", "2", "2F"]:
      for t in [ "Hex", "Tet" ]:
         for c in ["Yes", "No"]:
            for r in ["Yes", "No"]:
               filename="mesh_D%s_o%s_T%s_Contacts%s_Rich%s.fly"%(d,o,t,c,r)
               # certain cases are excluded:
               if  ( o == "2F" and t == "Tet" ) or \
                   ( t == "Tet" and r == "Yes" ) or \
                   ( o == "2F" and r == "Yes" )  :
                 pass
               else:
                  print("generate file ",filename)
                  if c == "Yes":
                     NE_X=int(NE**(1./d)/2+0.5)
                     NE_Y=int(NE**(1./d)+0.5)
                  else:
                     NE_X=int(NE**(1./d)+0.5)
                     NE_Y=NE_X
                  print(filename)
                  print("generating ",NE_X, NE_Y)
                  if o == "2":
                     o2=2
                     full=False
                  elif o == "2F":
                     o2=2
                     full=True
                  else:
                     o2=1
                     full=False
                  if c == "Yes":
                     dom1=getMesh(NE_X, NE_Y,t,d,o2,full,r=="Yes",0.5)
                     dom2=getMesh(NE_X, NE_Y,t,d,o2,full,r=="Yes",0.5)
                     x=dom2.getX().copy()
                     x[0]=1.-x[0]
                     dom2.setX(x)
                     dom=JoinFaces([dom1,dom2])
                  else:
                     dom=getMesh(NE_X, NE_Y,t,d,o2,full,r=="Yes",1.)
                  # set tags:
                  for fs in [ContinuousFunction(dom), Function(dom), FunctionOnBoundary(dom), FunctionOnContactOne(dom)]:
                       m=whereNegative(fs.getX()[d-1]-CUT)
                       fs.getDomain().setTagMap('tag1',1)
                       fs.setTags('tag1',m)
                       fs.setTags(100,1-m)
                  dom.write(os.path.join(MESH_DIRECTORY,filename))
                  # saveVTK(os.path.join(MESH_DIRECTORY,filename+".vtu"),dom)




"""Tool-specific initialization for clang-mp vi macport.
"""

# Based on SCons/Tool/gcc.py by Paweł Tomulik 2014 as a separate tool.

import os
import re
import subprocess

import SCons.Util
import SCons.Tool.cc
from SCons.Tool.clangCommon import get_clang_install_dirs
from SCons.Tool.MSCommon import msvc_setup_env_once

def generate(env, **kwargs):
    """Add Builders and construction variables for clang to an Environment."""
    compilers = ['clang++-mp-' + kwargs.get('version', '10')]
    SCons.Tool.cc.generate(env)
    if env['PLATFORM'] == 'win32':
        # Ensure that we have a proper path for clang
        clang = SCons.Tool.find_program_path(env, compilers[0], 
                                             default_paths=get_clang_install_dirs(env['PLATFORM']))
        if clang:
            clang_bin_dir = os.path.dirname(clang)
            env.AppendENVPath('PATH', clang_bin_dir)

            # Set-up ms tools paths
            msvc_setup_env_once(env)


    #print(env.Detect(compilers), compilers, env['CLANGMP_VERSION'])
    env['CXX'] = env.Detect(compilers) or compilers[0]
    if env['PLATFORM'] in ['cygwin', 'win32']:
        env['SHCXXFLAGS'] = SCons.Util.CLVar('$CXXFLAGS')
    else:
        env['SHCXXFLAGS'] = SCons.Util.CLVar('$CXXFLAGS -fPIC')

    # determine compiler version
    if env['CXX']:
        #pipe = SCons.Action._subproc(env, [env['CC'], '-dumpversion'],
        pipe = SCons.Action._subproc(env, [env['CXX'], '--version'],
                                     stdin='devnull',
                                     stderr='devnull',
                                     stdout=subprocess.PIPE)
        if pipe.wait() != 0: return
        # clang -dumpversion is of no use
        with pipe.stdout:
            line = pipe.stdout.readline()
        line = line.decode()
        match = re.search(r'clang +version +([0-9]+(?:\.[0-9]+)+)', line)
        if match:
            env['CXXVERSION'] = match.group(1)

def exists(env):
    return env.Detect(compilers)




##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
# See CREDITS file for contributors and development history
#
##############################################################################





__copyright__="""Copyright (c) 2003-2026 by the esys.escript Group
https://github.com/LutzGross/esys-escript.github.io
Primary Business: Queensland, Australia"""
__license__="""Licensed under the Apache License, version 2.0
http://www.apache.org/licenses/LICENSE-2.0"""
__url__="https://github.com/LutzGross/esys-escript.github.io"

from esys.escript.minimizer import MinimizerLBFGS, CostFunction, MinimizerException
import esys.escriptcore.utestselect as unittest
from esys.escriptcore.testing import *
import sys
import numpy as np

def createTest(p=2, dimension=3, numPoints=30):
    """
    this creates a simple test to find the point with the smallest distance from set of data points
    in the l^{p}-norm
    """
    import numpy as np
    from scipy.optimize import minimize_scalar

    theData = np.random.randint(-10, 10 + 1, size=(dimension, numPoints)).astype(float)
    def F0(x, *args):
        return np.sum(abs(x - theData[args[0], :]) ** p)

    x_true = np.zeros((dimension,))
    for i in range(dimension):
        R = minimize_scalar(F0, args=(i,), tol=1e-10)
        x_true[i] = R.x
    print("p = %g" % p)
    print("data = %s" % np.array_str(theData, precision=15, max_line_width=300))
    print("m_true = %s" % np.array_str(x_true, precision=15))
    return p, theData, x_true

class MinDist(CostFunction):
        def __init__(self, p=2, data=[], m_true=[]):
            super().__init__()

            self.p=p
            self.data=np.array(data)
            self.m_true = np.array(m_true)
            assert self.m_true.shape[0] == self.data.shape[0]
            self.dimension = self.m_true.shape[0]

        def getDualProduct(self, m, g):
            return np.dot(m, g)

        def getValue(self, m, *args):
            d,n = self.data.shape
            out=0
            for i in range(n):
                out+=np.sum(abs(m-self.data[:,i])**self.p)
            return out

        def getGradient(self, m, *args):
            d, n = self.data.shape
            out=np.zeros((d, ))
            for i in range(n):
                out+=self.p*abs(m-self.data[:,i])**(self.p-1)*np.sign(m-self.data[:,i])
            return out

        def getNorm(self, m):
            return np.linalg.norm(m)

        def getInverseHessianApproximation(self, r, m, *args, initializeHessian=False):
            if initializeHessian :
                d, n = self.data.shape
                self.h=np.zeros((d, ))
                for i in range(n):
                    self.h+=abs(m-self.data[:,i])**(self.p-2) #*self.p*(self.p-1)
            s = np.linalg.solve(np.diag(self.h), r)
            return s

class Test_MinimizerBFGS(unittest.TestCase):
    TEST1 = { 'p' : 2.5,
              'data': [[  5.,    3.,   -2.,   -8.,    0.,  -10.,   -3.,    8.,   -7.,    1.,   -7.,    4.,    1.,   10.,  -10.,   -1.,    3.,    6.,   -7.,    3.,    9.,    0.,    2.,   -8.,    4.,    7.,    1.,    4.,    5.,   -7., ],
                     [  2.,   -2.,    3.,   10.,    2.,   -3.,   -2.,   -9.,  -10.,   -8.,    4.,   10.,   -6.,    3.,   -3.,    9.,   -2.,   -6.,   -6.,    8.,    5.,    9.,   -9.,    7.,    6.,    3.,    6.,    5.,    8.,   -1., ],
                     [  9.,    8.,    5.,    9.,   -8.,    5.,    7.,    7.,   -2.,    4.,   -5.,   -5.,  -10.,    3.,   -7.,    9.,    2.,   -1.,   -3.,   -4.,    3.,   -3.,    9.,   10.,  -10.,   -5.,    0.,   -9.,    6.,    4., ]] ,
             'm_true' : [-0.101614684980357,  0.884081483782368,  0.759171570197073] }

    TEST2 = {'p' : 5,
            'data': [[-3., -10.,   7.,   3., -10., -5., 6., -1.,  10., -7.,  6., -1.,  6., 8., -8., 3., 3., -3., -2., -1., 10., -7., 6.,  5.,  1.,  0.,  8.,  5., -1., 8.],
                        [9., -7., -1., 3., 7., 7., -10., -2., -7., 5., -7., -3., -6., -2., -8.,  10., 4., 2., -10., -1., 10., 2., -7., -7., -10., -5., 0., -4., 3.,  10.],
                        [-4.,  5., -3., -1., 7., -4., 1., -4., -3., 8., -7., -3., -5., 0., -7., -1., 1., -5., 4., 10., 0., -4., 7., 3., -5., 4., -6., -1., -8., 5.]],
            'm_true':  [0.349084852003848, -0.144257203323146,  0.388851745831033] }

    TEST3 = { 'p':  9,
            'data': [[7., 2., -8., 9.,  2., 4., -4., 10., 3., 1., -4., 8., -5., -9., -8., 6., 10., -7., -7., 4., -3., -5., 1.,   10.,    1.,  -7.,    6.,    7.,    5.,  -5.],
                    [5.,  - 10.,  - 5.,    3.,  - 10.,    4.,  - 3.,  - 6.,  - 8.,  - 7.,    8.,    4.,  - 3.,    7.,  - 8.,  - 3.,  - 7.,    5.,   10.,    1.,  - 1., 9.,    5.,  - 6.,    8.,  - 3.,    6.,  - 4.,  - 9.,    8.],
                    [-3., 5., 10., 9., 9., 0., 5.,  - 4.,  - 3., 5.,  - 5.,  - 1., 0.,  - 6.,  - 7., 3.,  - 7.,  - 3., 10., 5., 7.,  - 8., 0., 7.,  - 5., 2.,  - 6., 10., 0., 7.]],
            'm_true': [0.756394642236015, -0.210685240100182,  1.315069006925737] }
    def testTEST1(self):
        F = MinDist(**self.TEST1)
        solve = MinimizerLBFGS(F, m_tol=1e-7, grad_tol=None, iterMax=100)
        m0 = np.zeros((F.dimension,))
        solve.run(m0)
        m = solve.getResult()
        self.assertTrue(np.allclose(m, F.m_true))

    def testTEST2(self):
        F = MinDist(**self.TEST2)
        solve = MinimizerLBFGS(F, m_tol=1e-7, grad_tol=None,  iterMax=100)
        m0 = np.zeros((F.dimension,))
        solve.run(m0)
        m = solve.getResult()
        self.assertTrue(np.allclose(m, F.m_true))

    def testTEST3(self):
        F = MinDist(**self.TEST3)
        solve = MinimizerLBFGS(F, m_tol=1e-7, grad_tol=None,  iterMax=100)
        m0 = np.zeros((F.dimension,))
        solve.run(m0)
        m = solve.getResult()
        self.assertTrue(np.allclose(m, F.m_true))

    def testTEST4(self):
        F = MinDist(**self.TEST1)
        solve = MinimizerLBFGS(F, m_tol=1e-10, grad_tol=1e-8, iterMax=100)
        m0 = np.zeros((F.dimension,))
        solve.run(m0)
        m = solve.getResult()
        #print((m-F.m_true)/F.m_true)
        self.assertTrue(np.allclose(m, F.m_true, rtol=1e-3))

    def testOptions(self):
        """
        test for setting options
        """
        F = MinDist(**self.TEST1)
        solve = MinimizerLBFGS(F, m_tol=1e-2, iterMax=23)
        self.assertEqual(solve.getOptions()['m_tol'], 1e-2)
        self.assertEqual(solve.getOptions()['iterMax'], 23)

        solve.setTolerance(m_tol=2., grad_tol=1.)
        solve.setIterMax(iterMax=1234)
        self.assertEqual(solve.getOptions()['m_tol'], 2.)
        self.assertEqual(solve.getOptions()['grad_tol'], 1.)
        self.assertEqual(solve.getOptions()['iterMax'], 1234)
        # test options:
        self.assertRaises(KeyError, solve.setOptions, BOB=-1)

        OPTS={'m_tol' :1,
        'grad_tol' : 2,
        'iterMax' : 3,
        'truncation' : 4,
        'restart' : 5,
        'relAlphaMin' : 6,
        'scaleSearchDirection' : 7,
        'initialAlpha' : 8}

        solve.setOptions(**OPTS)
        newOPTS=solve.getOptions()
        self.assertTrue('line_search' in newOPTS.keys())
        del newOPTS['line_search']
        self.assertDictEqual(newOPTS, OPTS)

        # linesearch:
        self.assertRaises(KeyError, solve.getLineSearch().setOptions, BOB=-1)
        lsOPTS={
        'phiEpsilon': 1,
        'alphaMin': 3,
        'alphaMax': 4,
        'overStepFactor': 5,
        'iterMax': 6,
        'c1': 7,
        'c2': 8,
        'inter_order': 2,
        'inter_tol': 9,
        'inter_iterMax': 10,
        'alphaOffset': 11,
        'alphaWidthMin': 12,
        'zoom_iterMax': 13,
        'zoom_reductionMin': 0.5
        }
        solve.getLineSearch().setOptions(**lsOPTS)
        newOPTS=solve.getOptions()
        self.assertDictEqual(newOPTS['line_search'], lsOPTS)


        # test of call back is called:
        def CB(iterCount, m, dm, Fm, grad_Fm, norm_m, args_m, failed):
            raise ValueError
        self.assertRaises(TypeError, solve.setCallback, 5)
        solve.setCallback(CB)
        self.assertRaises(ValueError, solve.run, np.zeros((F.dimension,)))

    def testRESTART(self):
        """
        Verify that a small restart interval still converges to the correct
        solution.  With restart=3 the L-BFGS history is flushed every 3 steps,
        exercising the scheduled-restart path.
        """
        F = MinDist(**self.TEST3)
        solve = MinimizerLBFGS(F, m_tol=1e-7, grad_tol=None, iterMax=200)
        solve.setOptions(restart=3)
        m0 = np.zeros((F.dimension,))
        solve.run(m0)
        m = solve.getResult()
        self.assertTrue(np.allclose(m, F.m_true))

if __name__ == '__main__':
    run_tests(__name__, exit_on_failure=True)





##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
# See CREDITS file for contributors and development history
#
##############################################################################

"""
A small set of functions to automate test discovery and running and also 
allow specific tests to be run without modifying the test scripts themselves.

When imported into a test file, any external script can import the test file
module and run tests as they wish.

Both specific and general forms of the functions return the result state,
allowing access to list all skipped tests and other data kept by the 
`TestTextResults` returned.

Examples:

Running all test classes::

    import your_test_module_or_file as tests
    tests.run_tests("your_test_module_or_file")

Running specific test classes::

    import your_test_module_or_file as tests
    tests.run_tests("your_test_module_or_file", [tests.Test_classA, tests.Test_classB])

Running a specific test within a class::

    import your_test_module_or_file as tests
    tests.run_single_test(tests.Test_classname("test_functionname"))

Printing the list of skipped tests::

    results = tests.run_tests("name")
    for skipped_test in results.skipped:
        print(skipped_test)

"""


__copyright__="""Copyright (c) 2003-2026 by the esys.escript Group
https://github.com/LutzGross/esys-escript.github.io
Primary Business: Queensland, Australia"""
__license__="""Licensed under the Apache License, version 2.0
http://www.apache.org/licenses/LICENSE-2.0"""
__url__="https://github.com/LutzGross/esys-escript.github.io"

import esys.escriptcore.utestselect as unittest
from .util import getMPIRankWorld
from .escriptcpp import MPIBarrierWorld
import os
import sys

def __add_tests(suite, modulename):
    """
    Discovers and adds all test classes from a module to a test suite.

    Test classes are identified by having a name starting with "Test" and
    being a subclass of ``unittest.TestCase``.

    :param suite: the test suite to add tests to
    :type suite: ``unittest.TestSuite``
    :param modulename: name of the module to discover tests from
    :type modulename: ``str``
    """
    import inspect
    clsmembers = inspect.getmembers(sys.modules[modulename], inspect.isclass)
    for name, cls in clsmembers:
        if modulename == cls.__module__ and name.startswith("Test") \
                and issubclass(cls, unittest.TestCase):
            suite.addTest(unittest.TestLoader().loadTestsFromTestCase(cls))

def run_single_test(test, exit_on_failure=False):
    """
    Runs a single test case and returns the result.

    :param test: the test instance to run
    :type test: ``unittest.TestCase``
    :param exit_on_failure: if ``True``, exits with code 1 on test failure
    :type exit_on_failure: ``bool``
    :return: the test result object containing pass/fail information
    :rtype: ``unittest.TestResult``
    """
    suite = unittest.TestSuite()
    suite.addTest(test)
    s=unittest.TextTestRunner(verbosity=2).run(suite)
    if exit_on_failure and not s.wasSuccessful():
        sys.exit(1)
    return s

def run_tests(modulename, classes = [], exit_on_failure = False):
    """
    Runs test classes from a module and returns the result.

    If no specific classes are provided, all test classes (those starting
    with "Test" and subclassing ``unittest.TestCase``) are discovered and
    run automatically.

    In MPI environments, only rank 0 produces output; other ranks run
    silently.

    :param modulename: name of the module containing the tests
    :type modulename: ``str``
    :param classes: specific test classes to run; if empty, all tests
                    in the module are discovered and run
    :type classes: ``list`` of ``unittest.TestCase`` subclasses
    :param exit_on_failure: if ``True``, exits with code 1 on test failure
    :type exit_on_failure: ``bool``
    :return: the test result object containing pass/fail information
    :rtype: ``unittest.TestResult``
    """
    rank = getMPIRankWorld()
    stream=sys.stderr #default
    verb=2
    if rank > 0:
        stream=open(os.devnull,'w')
        verb=0

    suite = unittest.TestSuite()
    if len(classes) == 0:
        __add_tests(suite, modulename)
    else:
        for test_class in classes:
            suite.addTest(unittest.TestLoader().loadTestsFromTestCase(test_class))
    s=unittest.TextTestRunner(stream=stream,verbosity=verb).run(suite)
    if exit_on_failure and not s.wasSuccessful():
        sys.stderr.flush()
        MPIBarrierWorld()
        sys.exit(1)
    return s




##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
# See CREDITS file for contributors and development history
#
##############################################################################

__copyright__="""Copyright (c) 2003-2026 by the esys.escript Group
https://github.com/LutzGross/esys-escript.github.io
Primary Business: Queensland, Australia"""
__license__="""Licensed under the Apache License, version 2.0
http://www.apache.org/licenses/LICENSE-2.0"""
__url__="https://github.com/LutzGross/esys-escript.github.io"

from esys.escript import *
from esys.finley import Rectangle
from esys.escript.pdetools import Locator
from esys.escript.linearPDEs import LinearSinglePDE
import numpy as np
import matplotlib.pyplot as plt

L0 = 80000   # horizontal extend [m]
L1 = 40000   # depth [m]
NE0 = 400    # number of element in horizontal direction
NE1 = 200    # number of element in vertical direction

rho_b = 100  # background resistivity [Ohm m]
rho_a = 0.5  # resistivity in the anomaly [Ohm m]
D = 250.     # depth of the top edge of the anomaly [m]
W = 1000.    # width of the anomaly [m]
H = 2000.    # vertical extend of the anomaly [m]
OFFSET = 1000 # offset of the central vertical axis of the anomaly from the domain (<0 move to the left, 0> move to the right)
f = 1.       # frequency [Hz]


Mu0=4*np.pi*1e-7
# ... mesh resolution:
h0, h1=L0/NE0, L1/NE1

mkDir("output")
# create Domain:
domain = Rectangle(n0=NE0, n1=NE1, l0=L0, l1=L1)

# set-up PDE. This one is complex:"
pde = LinearSinglePDE(domain, isComplex=True)
pde.setValue(D=1j*2*np.pi*f*Mu0)

# mask for anomaly boundaries:
X = ReducedFunction(domain).getX()
m1 = whereNonPositive(X[1]-(L1-D))
m2 = whereNonNegative(X[1]-(L1-D-H))
m3 = whereNonNegative(X[0]-(L0/2+OFFSET-W/2))
m4 = whereNonPositive(X[0]-(L0/2+OFFSET+W/2))
#  product of the masks defines mask for the anomaly:
m = m1*m2*m3*m4
# ... create control plot:
m_np = convertToNumpy(m)
x_np = convertToNumpy(m.getX())

plt.figure()
plt.tricontourf(x_np[0], x_np[1], m_np[0], 15)
plt.xlabel('x [m]')
plt.ylabel('y [m]')
plt.colorbar()
plt.savefig('output/mt_anomaly.png')
print("anomaly plot generated.")
# now we can set the resistivity distribution:
rho = rho_b*(1-m)+rho_a*m
rho = interpolate(rho, Function(domain))
pde.setValue(A=rho*np.eye(2))
# set magnetic field to one at the top of the domain:
x = domain.getX()
mD = whereZero(x[1]-L1)
print("SS")
pde.setValue(q=mD, r=1)
# now we can get the solution:
Hx = pde.getSolution()
print("SS")
# electrical field:
Ey = rho*grad(Hx, ReducedFunction(domain))[1]
# impedance
Zyx = Ey/Hx
print("SS")
# =======================================
#
# .. we grab the impedance along a transect
#
locations_in_transect = [( h0*k+h0/2. , L1 ) for k in range(0, NE0, 2)]
locator_transect = Locator(where=ReducedFunction(domain), x=locations_in_transect)
#
# .. horizontal positions and Z_yx along transect
#
x_ts = np.array([ x[0] for x in locator_transect.getX() ])
Zyx_ts=np.array(locator_transect(Zyx))
#
# and save to file:
#
np.savetxt('output/MTData2.csv', (x_ts, Zyx_ts.real, Zyx_ts.imag ), delimiter=',')
#
# ... plot the apparent resistivity and phase:
#
rho_a_ts = 1./(2*np.pi*f*Mu0)*abs(Zyx_ts)**2
phi_ts = np.angle(Zyx_ts, deg=True)
#
# and then we can plot these data:
#
plt.figure()
plt.plot(x_ts, rho_a_ts)
plt.xlabel('x [m]')
plt.ylabel('resistivity [m]')
plt.title(f"Resistivity vs offset  (f={f} Hz)")
plt.savefig('output/mt_resistivity.png')
print("resistivity plot generated.")

plt.figure()
plt.plot(x_ts, phi_ts)
plt.xlabel('x [m]')
plt.ylabel('phase [deg]')
plt.title(f"phase vs offset (f={f} Hz)")
plt.savefig("output/mt_phase.png")
print("phase plot generated.")




##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
# See CREDITS file for contributors and development history
#
##############################################################################

# No-MPI configuration for Linux Mint.
# Imports the full mint_options.py and disables MPI.

from scons.templates.mint_options import *

# Disable MPI
mpi = 'none'
mpi4py = False

# ParMETIS requires MPI
parmetis = False




########################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# Earth Systems Science Computational Center (ESSCC)
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
########################################################


__copyright__="""Copyright (c) 2003-2026 by the esys.escript Group
Earth Systems Science Computational Center (ESSCC)
https://github.com/LutzGross/esys-escript.github.io
Primary Business: Queensland, Australia"""
__license__="""Licensed under the Apache License, version 2.0
http://www.apache.org/licenses/LICENSE-2.0"""
__url__="https://github.com/LutzGross/esys-escript.github.io"

"""
Test suite for the linearPDE  and pdetools test on finley

:remark:

:var __author__: name of author
:var __licence__: licence agreement
:var __url__: url entry point on documentation
:var __version__: version
:var __date__: date of the version
"""

__author__="Lutz Gross, l.gross@uq.edu.au"

import esys.escriptcore.utestselect as unittest
from esys.escriptcore.testing import *
from test_nonLinearPDE import Test_nlpde
from esys.finley import Rectangle, Brick

class Test_nonLinearPDEOnFinley2D(Test_nlpde):
   def setUp(self):
        self.domain = Rectangle(l0=1.,l1=1.,n0=10, n1=10) 
   def tearDown(self):
        del self.domain

class Test_nonLinearPDEOnFinley3D(Test_nlpde):
   def setUp(self):
        self.domain = Brick(l0=1.,l1=1.,l2=1.,n0=10, n1=10,n2=10) 
   def tearDown(self):
        del self.domain

if __name__ == '__main__':
    run_tests(__name__, exit_on_failure=True)

